/*====================================================================*\

App.java

Application class.

------------------------------------------------------------------------

This file is part of NlfExample, a Java application that demonstrates
the use of the uk.org.blankaspect.nlf API.
Copyright 2007-2013 Andy Morgan-Richards.

NlfExample is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// IMPORTS


import exception.AppException;
import exception.TaskCancelledException;

import java.io.File;
import java.io.IOException;

import java.util.ArrayList;
import java.util.List;

import util.CalendarTime;
import util.CommandLine;
import util.NoYes;
import util.PropertyString;
import util.ResourceProperties;

//----------------------------------------------------------------------


// APPLICATION CLASS


class App
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    public static final     String  SHORT_NAME  = "NlfExample";
    public static final     String  LONG_NAME   = "Nested List File example";
    public static final     String  NAME_KEY    = "nlfExample";

    private static final    String  DEBUG_PROPERTY_KEY      = "app.debug";
    private static final    String  VERSION_PROPERTY_KEY    = "version";
    private static final    String  BUILD_PROPERTY_KEY      = "build";
    private static final    String  RELEASE_PROPERTY_KEY    = "release";

    private static final    String  BUILD_PROPERTIES_PATHNAME   = "resources/build.properties";

    private static final    String  DEBUG_STR   = " Debug";

    private static final    String  USAGE_STR   =
        "Usage: " + NAME_KEY + " [options] command\n" +
        "\n" +
        "Commands:\n" +
        "  --generate\n" +
        "      Generate a Nested List File, output file and, optionally, an XML file.\n" +
        "  --help\n" +
        "      Display help information.\n" +
        "  --version\n" +
        "      Display version information.\n" +
        "\n" +
        "Options:\n" +
        "  --nlf=pathname\n" +
        "      The pathname of the Nested List File (NLF) that will be generated.\n" +
        "      If no NLF pathname is specified, a temporary file is created for the NLF.\n" +
        "      The temporary file is deleted when the program terminates.\n" +
        "  --output=pathname\n" +
        "      The pathname of the output file.\n" +
        "      If no output pathname is specified, the output is written to standard\n" +
        "      output.\n" +
        "  --overwrite\n" +
        "      Overwrite an existing file without prompting.\n" +
        "  --verbose\n" +
        "      Write full information to standard output.\n" +
        "  --xml=pathname\n" +
        "      The pathname of the XML file that will be generated from the NLF.\n" +
        "      If no XML pathname is specified, no XML file is generated.\n" +
        "\n" +
        "If an option takes an argument, the key and argument of the option may be\n" +
        "separated either by whitespace or by a single \"=\".";

    private static final    String  READING_STR     = "Reading ";
    private static final    String  WRITING_STR     = "Writing ";
    private static final    String  REPLACE_STR     = " already exists.\nDo you want to replace it?";
    private static final    String  REPLACE_PROMPT  = "Replace or Quit (R/Q)";

    private enum Command
    {
        GENERATE,
        HELP,
        VERSION
    }

    private interface ExitCode
    {
        int SUCCESS = 0;
        int ERROR   = 1;
    }

////////////////////////////////////////////////////////////////////////
//  Enumerated types
////////////////////////////////////////////////////////////////////////


    // COMMAND-LINE OPTIONS


    private enum Option
        implements CommandLine.Option<Option>
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        GENERATE    ( "generate",   false ),
        HELP        ( "help",       false ),
        NLF         ( "nlf",        true ),
        OUTPUT      ( "output",     true ),
        OVERWRITE   ( "overwrite",  false ),
        VERBOSE     ( "verbose",    false ),
        VERSION     ( "version",    false ),
        XML         ( "xml",        true );

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        private Option( String  name,
                        boolean hasArgument )
        {
            this.name = name;
            this.hasArgument = hasArgument;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : CommandLine.Option interface
    ////////////////////////////////////////////////////////////////////

        public Option getKey( )
        {
            return this;
        }

        //--------------------------------------------------------------

        public String getName( )
        {
            return name;
        }

        //--------------------------------------------------------------

        public boolean hasArgument( )
        {
            return hasArgument;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  name;
        private boolean hasArgument;

    }

    //==================================================================


    // ERROR IDENTIFIERS


    private enum ErrorId
        implements AppException.Id
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        CONFLICTING_OPTION_ARGUMENTS
        ( "The %1 option was specified more than once with different arguments." ),

        FAILED_TO_CREATE_TEMPORARY_FILE
        ( "Failed to create a temporary file." ),

        NO_COMMAND
        ( "No command was specified." ),

        MULTIPLE_COMMANDS
        ( "More than one command was specified." );

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        private ErrorId( String message )
        {
            this.message = message;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : AppException.Id interface
    ////////////////////////////////////////////////////////////////////

        public String getMessage( )
        {
            return message;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  message;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Member classes : non-inner classes
////////////////////////////////////////////////////////////////////////


    // COMMAND-LINE USAGE EXCEPTION CLASS


    private static class UsageException
        extends AppException
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        private UsageException( ErrorId id )
        {
            super( id );
            suffix = "\n" + USAGE_STR;
        }

        //--------------------------------------------------------------

        private UsageException( ErrorId   id,
                                String... strs )
        {
            super( id, strs );
            suffix = "\n" + USAGE_STR;
        }

        //--------------------------------------------------------------

        private UsageException( ErrorId                     id,
                                CommandLine.Element<Option> element )
        {
            this( id );
            prefix = element.getOptionString( ) + " " + element.getValue( ) + "\n";
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : overriding methods
    ////////////////////////////////////////////////////////////////////

        @Override
        protected String getPrefix( )
        {
            return prefix;
        }

        //--------------------------------------------------------------

        @Override
        protected String getSuffix( )
        {
            return suffix;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  prefix;
        private String  suffix;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    private App( )
    {
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    public static void main( String[] args )
    {
        System.exit( getInstance( ).init( args ) );
    }

    //------------------------------------------------------------------

    public static App getInstance( )
    {
        if ( instance == null )
            instance = new App( );
        return instance;
    }

    //------------------------------------------------------------------

    public static boolean isDebug( )
    {
        return debug;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    public String getVersionString( )
    {
        StringBuilder buffer = new StringBuilder( 32 );
        String str = buildProperties.get( VERSION_PROPERTY_KEY );
        if ( str != null )
            buffer.append( str );

        str = buildProperties.get( RELEASE_PROPERTY_KEY );
        if ( str == null )
        {
            long time = System.currentTimeMillis( );
            if ( buffer.length( ) > 0 )
                buffer.append( ' ' );
            buffer.append( 'b' );
            buffer.append( CalendarTime.dateToString( time ) );
            buffer.append( '-' );
            buffer.append( CalendarTime.hoursMinsToString( time ) );
        }
        else
        {
            NoYes release = NoYes.get( str );
            if ( (release == null) || !release.toBoolean( ) )
            {
                str = buildProperties.get( BUILD_PROPERTY_KEY );
                if ( str != null )
                {
                    if ( buffer.length( ) > 0 )
                        buffer.append( ' ' );
                    buffer.append( str );
                }
            }
        }

        if ( debug )
            buffer.append( DEBUG_STR );

        return buffer.toString( );
    }

    //------------------------------------------------------------------

    private int init( String[] arguments )
    {
        // Assume success
        int exitCode = ExitCode.SUCCESS;

        // Set runtime debug flag
        debug = (System.getProperty( DEBUG_PROPERTY_KEY ) != null);

        // Read build properties
        buildProperties = new ResourceProperties( BUILD_PROPERTIES_PATHNAME, getClass( ) );

        // Parse command line
        try
        {
            List<Command> commands = new ArrayList<Command>( );
            boolean verbose = false;
            boolean overwrite = false;

            for ( CommandLine.Element<Option> element :
                            new CommandLine<Option>( Option.class, true, USAGE_STR ).parse( arguments ) )
            {
                String elementValue = element.getValue( );

                switch ( element.getOption( ).getKey( ) )
                {
                    case GENERATE:
                        if ( !commands.contains( Command.GENERATE ) )
                            commands.add( Command.GENERATE );
                        break;

                    case HELP:
                        if ( !commands.contains( Command.HELP ) )
                            commands.add( Command.HELP );
                        break;

                    case NLF:
                    {
                        String nlfPathname0 = PropertyString.parsePathname( elementValue );
                        if ( (nlfPathname != null) && !nlfPathname.equals( nlfPathname0 ) )
                            throw new UsageException( ErrorId.CONFLICTING_OPTION_ARGUMENTS,
                                                      element.getOptionString( ) );
                        nlfPathname = nlfPathname0;
                        break;
                    }

                    case OUTPUT:
                    {
                        String outPathname0 = PropertyString.parsePathname( elementValue );
                        if ( (outPathname != null) && !outPathname.equals( outPathname0 ) )
                            throw new UsageException( ErrorId.CONFLICTING_OPTION_ARGUMENTS,
                                                      element.getOptionString( ) );
                        outPathname = outPathname0;
                        break;
                    }

                    case OVERWRITE:
                        overwrite = true;
                        break;

                    case VERBOSE:
                        verbose = true;
                        break;

                    case VERSION:
                        if ( !commands.contains( Command.VERSION ) )
                            commands.add( Command.VERSION );
                        break;

                    case XML:
                    {
                        String xmlPathname0 = PropertyString.parsePathname( elementValue );
                        if ( (xmlPathname != null) && !xmlPathname.equals( xmlPathname0 ) )
                            throw new UsageException( ErrorId.CONFLICTING_OPTION_ARGUMENTS,
                                                      element.getOptionString( ) );
                        xmlPathname = xmlPathname0;
                        break;
                    }
                }
            }

            // Test for commands
            if ( commands.isEmpty( ) )
                throw new UsageException( ErrorId.NO_COMMAND );
            if ( commands.size( ) > 1 )
                throw new UsageException( ErrorId.MULTIPLE_COMMANDS );

            // Perform command
            Command command = commands.get( 0 );
            switch ( command )
            {
                case GENERATE:
                    if ( verbose )
                        showTitle( );
                    onGenerate( overwrite, verbose );
                    break;

                case HELP:
                    showTitle( );
                    System.out.println( );
                    System.out.println( USAGE_STR );
                    break;

                case VERSION:
                    showTitle( );
                    break;
            }

        }
        catch ( TaskCancelledException e )
        {
            // ignore
        }
        catch ( AppException e )
        {
            if ( !titleShown )
                System.err.println( SHORT_NAME + " " + getVersionString( ) );
            System.err.println( e );
            exitCode = ExitCode.ERROR;
        }

        return exitCode;
    }

    //------------------------------------------------------------------

    private void showTitle( )
    {
        if ( !titleShown )
        {
            System.out.println( SHORT_NAME + " " + getVersionString( ) );
            titleShown = true;
        }
    }

    //------------------------------------------------------------------

    private void confirmAction( String promptStr,
                                char   approveChar,
                                char   cancelChar )
        throws TaskCancelledException
    {
        while ( true )
        {
            try
            {
                while ( System.in.available( ) > 0 )
                    System.in.read( );
                System.out.print( promptStr + " ? " );
                char ch = Character.toUpperCase( (char)System.in.read( ) );
                if ( ch == Character.toUpperCase( approveChar ) )
                    break;
                if ( ch == Character.toUpperCase( cancelChar ) )
                    throw new TaskCancelledException( );
            }
            catch ( IOException e )
            {
                // ignore
            }
        }
    }

    //------------------------------------------------------------------

    private void onGenerate( boolean overwrite,
                             boolean verbose )
        throws AppException
    {
        File nlfFile = null;
        boolean deleteNlf = false;
        try
        {
            // Create File object for NLF
            if ( nlfPathname == null )
            {
                try
                {
                    nlfFile = File.createTempFile( AppConstants.TEMP_FILE_PREFIX, null );
                    deleteNlf = true;
                }
                catch ( Exception e )
                {
                    throw new AppException( ErrorId.FAILED_TO_CREATE_TEMPORARY_FILE, e );
                }
            }
            else
                nlfFile = new File( nlfPathname );

            // Create File object for output file; prompt to replace an existing file
            File outFile = null;
            if ( outPathname != null )
            {
                outFile = new File( outPathname );
                if ( !overwrite && outFile.isFile( ) )
                {
                    System.out.println( Util.getPathname( outFile ) + REPLACE_STR );
                    confirmAction( REPLACE_PROMPT, 'R', 'Q' );
                }
            }

            // Create File object for XML file; prompt to replace an existing file
            File xmlFile = null;
            if ( xmlPathname != null )
            {
                xmlFile = new File( xmlPathname );
                if ( !overwrite && xmlFile.isFile( ) )
                {
                    System.out.println( Util.getPathname( xmlFile ) + REPLACE_STR );
                    confirmAction( REPLACE_PROMPT, 'R', 'Q' );
                }
            }

            // Generate NLF file
            NlfProcessor nlfProcessor = new NlfProcessor( nlfFile );
            if ( verbose )
                System.out.println( WRITING_STR + Util.getPathname( nlfFile ) );
            nlfProcessor.write( );

            // Read NLF file and write its contents to output file
            if ( verbose )
                System.out.println( READING_STR + Util.getPathname( nlfFile ) );
            nlfProcessor.read( outFile, xmlFile );
        }
        catch ( AppException e )
        {
            // Rethrow exception
            throw e;
        }
        finally
        {
            // Delete temporary file
            if ( (nlfFile != null) && deleteNlf )
                nlfFile.delete( );
        }
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class variables
////////////////////////////////////////////////////////////////////////

    private static  App     instance;
    private static  boolean debug;

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    private ResourceProperties  buildProperties;
    private boolean             titleShown;
    private String              nlfPathname;
    private String              outPathname;
    private String              xmlPathname;

}

//----------------------------------------------------------------------
