/*====================================================================*\

NlfProcessor.java

Nested List File processor class.

------------------------------------------------------------------------

This file is part of NlfExample, a Java application that demonstrates
the use of the uk.org.blankaspect.nlf API.
Copyright 2007-2013 Andy Morgan-Richards.

NlfExample is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// IMPORTS


import exception.AppException;
import exception.FileException;

import java.io.DataOutput;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;

import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CoderResult;
import java.nio.charset.CodingErrorAction;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import java.util.zip.CRC32;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import uk.org.blankaspect.nlf.Attributes;
import uk.org.blankaspect.nlf.Chunk;
import uk.org.blankaspect.nlf.ChunkList;
import uk.org.blankaspect.nlf.Document;
import uk.org.blankaspect.nlf.Id;
import uk.org.blankaspect.nlf.NlfConstants;
import uk.org.blankaspect.nlf.NlfException;
import uk.org.blankaspect.nlf.NlfUtilities;

import util.NumberUtilities;
import util.StringUtilities;
import util.SystemUtilities;

import xml.Attribute;
import xml.XmlFile;
import xml.XmlUtilities;
import xml.XmlWriter;

//----------------------------------------------------------------------


// NESTED LIST FILE PROCESSOR CLASS


class NlfProcessor
    implements XmlFile.ElementWriter
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    private static final    int SEPARATOR_LENGTH        = 85;
    private static final    int INDENT_UNIT_WIDTH       = 4;
    private static final    int MAX_BINARY_DATA_LENGTH  = 16;

    private static final    String  ENCODING_NAME_UTF8  = "UTF-8";

    private static final    String  LIST_INSTANCE_ID_STR        = "List-instance ID";
    private static final    String  CHUNK_ID_STR                = "Chunk ID";
    private static final    String  ATTR_STR                    = "[attr]";
    private static final    String  LOCAL_NAMESPACE_NAME_STR    = "local namespace name";
    private static final    String  NAME_STR                    = "name";
    private static final    String  PATHNAME_STR                = "pathname";
    private static final    String  SIZE_STR                    = "size";
    private static final    String  DATA_STR                    = "data";
    private static final    String  HEX_STR                     = "(hex)";
    private static final    String  EQUALS_STR                  = " = ";

    private static final    String  OFFSET_STR  = "Offset = 0x";

    private static final    String  CRC_CORRECT_STR = "The CRC value is correct.";

    private static final    Id  ROOT_LIST_ID            = new Id( "NLF-Example" );
    private static final    Id  ALPHA_LIST_ID           = new Id( "Alpha" );
    private static final    Id  BETA_LIST_ID            = new Id( "Beta" );
    private static final    Id  GAMMA_LIST_ID           = new Id( "Gamma" );
    private static final    Id  DELTA_LIST_ID           = new Id( "Delta" );
    private static final    Id  CRC_ID                  = new Id( "crc" );
    private static final    Id  RANDOM_ID               = new Id( "random" );
    private static final    Id  BLAKE_ROSE_ID           = new Id( "blake-rose" );
    private static final    Id  BLAKE_TYGER_ID          = new Id( "blake-tyger" );
    private static final    Id  BLAKE_LONDON_ID         = new Id( "blake-london" );
    private static final    Id  BAUDELAIRE_SERPENT_ID   = new Id( "baudelaire-serpent" );
    private static final    Id  BAUDELAIRE_REVENANT_ID  = new Id( "baudelaire-revenant" );
    private static final    Id  BAUDELAIRE_VOYAGE_ID    = new Id( "baudelaire-voyage" );
    private static final    Id  WHITMAN_MYSELF_ID       = new Id( "whitman-myself" );
    private static final    Id  WHITMAN_FACES_ID        = new Id( "whitman-faces" );
    private static final    Id  WHITMAN_OSCEOLA_ID      = new Id( "whitman-osceola" );

    private static final    String  NS_NAME_EXAMPLE_STR
                                                = "http://ns.blankaspect.org.uk/example/nlf/example-1";
    private static final    String  NS_NAME_EUROPE_STR
                                                = "http://ns.blankaspect.org.uk/example/nlf/poet-europe";
    private static final    String  NS_NAME_USA_STR
                                                = "http://ns.blankaspect.org.uk/example/nlf/poet-usa";

    private static final    String  ENGLISH_UK_STR  = "en-GB";
    private static final    String  ENGLISH_USA_STR = "en-US";
    private static final    String  FRENCH_STR      = "fr-FR";

    private static final    String  BLAKE_ROSE_STR          = "O rose, thou art sick!";
    private static final    String  BLAKE_TYGER_STR         = "In what furnace was thy brain?";
    private static final    String  BLAKE_LONDON_STR        = "The mind-forg\u2019d manacles I hear";
    private static final    String  BAUDELAIRE_SERPENT_STR  = "Tes yeux, o\u00F9 rien ne se " +
                                                                "r\u00E9v\u00E8le";
    private static final    String  BAUDELAIRE_REVENANT_STR = "Comme les anges \u00E0 l\u2019\u0153il " +
                                                                "fauve";
    private static final    String  BAUDELAIRE_VOYAGE_STR   = "Amer savoir, celui qu\u2019on tire du " +
                                                                "voyage!";
    private static final    String  WHITMAN_MYSELF_STR      = "The friendly and flowing savage, who is he?";
    private static final    String  WHITMAN_FACES_STR       = "Snakes nest in that mouth";
    private static final    String  WHITMAN_OSCEOLA_STR     = "When his hour for death had come";

    private enum LineKind
    {
        NONE,
        VERTICAL,
        HORIZONTAL
    }

    private enum ChunkKind
    {
        NONE,
        CHUNK,
        LIST
    }

    private interface AttrName
    {
        String  LANGUAGE    = "language";
        String  TEXT        = "text";
    }

////////////////////////////////////////////////////////////////////////
//  Enumerated types
////////////////////////////////////////////////////////////////////////


    // ERROR IDENTIFIERS


    /**
     * This enumeration contains identifiers for the exceptions that may occur when reading, writing or
     * parsing a file.  Each identifier maps to a message.
     */

    private enum ErrorId
        implements AppException.Id
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        FAILED_TO_OPEN_FILE
        ( "Failed to open the file." ),

        FAILED_TO_CLOSE_FILE
        ( "Failed to close the file." ),

        ERROR_READING_FILE
        ( "An error occurred when reading the file." ),

        ERROR_WRITING_FILE
        ( "An error occurred when writing the file." ),

        FILE_ACCESS_NOT_PERMITTED
        ( "Access to the file was not permitted." ),

        CHUNK_DATA_ARE_TOO_LARGE
        ( "The chunk data are too large." ),

        INVALID_TEXT
        ( "The chunk data are not valid text." ),

        NO_CRC_CHUNK
        ( "The file does not have a CRC chunk." ),

        INCORRECT_CRC
        ( "The CRC value is incorrect." );

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs an exception identifier.
         *
         * @param message  the message that the exception identifier maps to.
         */

        private ErrorId( String message )
        {
            this.message = message;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : AppException.Id interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Returns the message that the exception identifier maps to.
         *
         * @return the message that the exception identifier maps to.
         */

        public String getMessage( )
        {
            return message;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  message;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Member classes : non-inner classes
////////////////////////////////////////////////////////////////////////


    // BINARY-CHUNK ENCODER CLASS


    /**
     * This singleton class encodes binary chunk data as a sequence of hexadecimal digit characters.
     */

    private static class BinaryChunkEncoder
        implements Chunk.Encoder
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a {@code BinaryChunkEncoder}.  This is a singleton class: only one instance of the
         * class is constructed, by {@link #getInstance( ) getInstance( )}.
         */

        private BinaryChunkEncoder( )
        {
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Class methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Returns the single instance of {@code BinaryChunkEncoder}.
         *
         * @return the single instance of {@code BinaryChunkEncoder}.
         */

        public static BinaryChunkEncoder getInstance( )
        {
            if ( instance == null )
                instance = new BinaryChunkEncoder( );
            return instance;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Encoder interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk encoder before any encoding is performed: does nothing.
         */

        public void reset( )
        {
            // do nothing
        }

        //--------------------------------------------------------------

        /**
         * Returns the length of a block of input data that the {@link #encode(byte[], int, int, boolean)
         * encode(byte[], int, int, boolean)} method expects to receive.
         *
         * @param  size  the size of the chunk data.
         * @return the length of a block of input data that the {@code encode(byte[], int, int, boolean)}
         *         method expects to receive.
         */

        public int getInputLength( long size )
        {
            return 0;
        }

        //--------------------------------------------------------------

        /**
         * Encodes the specified byte data as a sequence of hexadecimal digit characters.
         *
         * @param  data       the data that are to be encoded.
         * @param  offset     the offset to {@code data} at which the input data begin.
         * @param  length     the length of the input data.
         * @param  endOfInput {@code true} if the input data are the last that are to be encoded; {@code
         *                    false} otherwise.
         * @return the string that results from encoding the input data.
         * @throws IllegalArgumentException
         *           if
         *           <ul>
         *             <li>{@code data} is {@code null}, or</li>
         *             <li>{@code (offset < 0)} or {@code (offset > data.length)}, or</li>
         *             <li>{@code (length < 0)} or {@code (length > data.length - offset)}.</li>
         *           </ul>
         */

        public String encode( byte[]  data,
                              int     offset,
                              int     length,
                              boolean endOfInput )
            throws IllegalArgumentException
        {
            // Validate arguments
            if ( (data == null) || (offset < 0) || (offset > data.length) ||
                 (length < 0) || (length > data.length - offset) )
                throw new IllegalArgumentException( );

            // Encode input data
            StringBuffer buffer = new StringBuffer( length * 2 );
            int endOffset = offset + length;
            while ( offset < endOffset )
                buffer.append( NumberUtilities.byteToHexString( data[offset++] ) );
            return buffer.toString( );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Class variables
    ////////////////////////////////////////////////////////////////////

        private static  BinaryChunkEncoder  instance;

    }

    //==================================================================


    // TEXT-CHUNK ENCODER CLASS


    /**
     * This singleton class encodes chunk data as a Unicode string.  The input data are expected to be a
     * UTF-8 sequence, so the encoder's behaviour is more usually associated with a <em>decoder</em>.
     */

    private static class TextChunkEncoder
        implements Chunk.Encoder
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        private static final    int MAX_BLOCK_LENGTH    = 256;

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a {@code TextChunkEncoder}.  This is a singleton class: only one instance of the class
         * is constructed, by {@link #getInstance( ) getInstance( )}.
         */

        private TextChunkEncoder( )
        {
            decoder = Charset.forName( ENCODING_NAME_UTF8 ).newDecoder( ).
                                                        onMalformedInput( CodingErrorAction.REPORT ).
                                                        onUnmappableCharacter( CodingErrorAction.REPORT );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Class methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Returns the single instance of {@code TextChunkEncoder}.
         *
         * @return the single instance of {@code TextChunkEncoder}.
         */

        public static TextChunkEncoder getInstance( )
        {
            if ( instance == null )
                instance = new TextChunkEncoder( );
            return instance;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Encoder interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk encoder before any encoding is performed.
         */

        public void reset( )
        {
            decoder.reset( );
            inBuffer = null;
        }

        //--------------------------------------------------------------

        /**
         * Returns the length of a block of input data that the {@link #encode(byte[], int, int, boolean)
         * encode(byte[], int, int, boolean)} method expects to receive.
         *
         * @param  size  the size of the chunk data.
         * @return the length of a block of input data that the {@code encode(byte[], int, int, boolean)}
         *         method expects to receive.
         */

        public int getInputLength( long size )
        {
            return (int)Math.min( size, MAX_BLOCK_LENGTH );
        }

        //--------------------------------------------------------------

        /**
         * Encodes the specified byte data as a Unicode string.
         *
         * @param  data       the data that are to be encoded.
         * @param  offset     the offset to {@code data} at which the input data begin.
         * @param  length     the length of the input data.
         * @param  endOfInput {@code true} if the input data are the last that are to be encoded; {@code
         *                    false} otherwise.
         * @return the string that results from encoding the input data.
         * @throws IllegalArgumentException
         *           if
         *           <ul>
         *             <li>{@code data} is {@code null}, or</li>
         *             <li>{@code (offset < 0)} or {@code (offset > data.length)}, or</li>
         *             <li>{@code (length < 0)} or {@code (length > data.length - offset)}.</li>
         *           </ul>
         */

        public String encode( byte[]  data,
                              int     offset,
                              int     length,
                              boolean endOfInput )
            throws IllegalArgumentException
        {
            // Validate arguments
            if ( (data == null) || (offset < 0) || (offset > data.length) ||
                 (length < 0) || (length > data.length - offset) )
                throw new IllegalArgumentException( );

            // Allocate and initialise input buffer
            if ( inBuffer == null )
                inBuffer = ByteBuffer.allocate( length );
            else
            {
                ByteBuffer buffer = ByteBuffer.allocate( inBuffer.remaining( ) + length );
                buffer.put( inBuffer );
                inBuffer = buffer;
            }
            inBuffer.put( data, offset, length );
            inBuffer.flip( );

            // Decode input
            int outBufferSize = (length == 0) ? 1 : length;
            CharBuffer outBuffer = CharBuffer.allocate( outBufferSize );
            while ( true )
            {
                CoderResult result = decoder.decode( inBuffer, outBuffer, endOfInput );
                if ( result.isUnderflow( ) && endOfInput )
                    result = decoder.flush( outBuffer );
                if ( result.isUnderflow( ) )
                    break;
                if ( result.isOverflow( ) )
                {
                    outBufferSize *= 2;
                    CharBuffer buffer = CharBuffer.allocate( outBufferSize );
                    outBuffer.flip( );
                    buffer.put( outBuffer );
                    outBuffer = buffer;
                }
                else
                    throw new IllegalArgumentException( );
            }

            // Release input buffer
            if ( !inBuffer.hasRemaining( ) )
                inBuffer = null;

            // Return output
            outBuffer.flip( );
            return outBuffer.toString( );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Class variables
    ////////////////////////////////////////////////////////////////////

        private static  TextChunkEncoder    instance;

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private ByteBuffer      inBuffer;
        private CharsetDecoder  decoder;

    }

    //==================================================================


    // CHUNK PROCESSOR CLASS


    /**
     * This class processes the chunks that are visited in the traversal of the document tree of an NLF.
     * The {@link #process(uk.org.blankaspect.nlf.Chunk) process(uk.org.blankaspect.nlf.Chunk)} method sets
     * a chunk encoder on each chunk.  The type of encoder differs according to whether the chunk contains
     * text or binary data.
     */

    private static class ChunkProcessor
        implements Chunk.Processor
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a chunk processor that will set a chunk encoder on each chunk that is visited in the
         * traversal of the NLF document tree.
         */

        private ChunkProcessor( )
        {
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Processor interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Sets a chunk encoder on each chunk.  The type of encoder differs according to whether the chunk
         * contains text or binary data.
         *
         * @param chunk  the chunk that is to be processed.
         */

        public void process( Chunk chunk )
        {
            chunk.setEncoder( containsText( chunk ) ? TextChunkEncoder.getInstance( )
                                                    : BinaryChunkEncoder.getInstance( ) );
        }

        //--------------------------------------------------------------

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Member classes : inner classes
////////////////////////////////////////////////////////////////////////


    // CRC-CHUNK WRITER CLASS


    /**
     * This class implements a chunk writer for a CRC chunk.  It writes the CRC value of the data of the
     * other chunks in the NLF document.
     */

    private class CrcChunkWriter
        implements Chunk.Writer
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a chunk writer that will write the CRC value of the data of the other chunks in an
         * NLF.
         */

        private CrcChunkWriter( )
        {
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Writer interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk writer before any writing is performed.  The writer is reset at the start of
         * each pass of the document writer.
         *
         * @param  pass  the index of the pass (0 = first pass, 1 = second pass) of the document writer that
         *               calls this method.
         * @return {@code true} on the first pass to indicate that the chunk should be rewritten on the
         *         second pass.
         */

        public boolean reset( int pass )
        {
            return true;
        }

        //--------------------------------------------------------------

        /**
         * Returns the length (size) of the chunk in bytes.
         *
         * @return the length (size) of the chunk in bytes.
         */

        public long getLength( )
        {
            return ( Integer.SIZE >> 3 );
        }

        //--------------------------------------------------------------

        /**
         * Writes the chunk data to the specified data output, which is a {@code java.io.RandomAccessFile}.
         *
         * @param  dataOutput  the data output to which the chunk data are to be written.
         * @throws IOException
         *           if an I/O error occurs.
         */

        public void write( DataOutput dataOutput )
            throws IOException
        {
            dataOutput.writeInt( (int)crc.getValue( ) );
        }

        //--------------------------------------------------------------

    }

    //==================================================================


    // RANDOM-CHUNK WRITER CLASS


    /**
     * This class implements a chunk writer for a random-data chunk.  It generates a small amount of random
     * data, which it writes to a NLF document.
     * <p>
     * Because the size of the chunk is random, the size field in the chunk header must be fixed up after
     * the chunk has been written.  This is accomplished by having the {@link #getLength() getLength()}
     * method return {@code -1}.
     * </p>
     */

    private class RandomChunkWriter
        implements Chunk.Writer
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        private static final    int MIN_LENGTH  = 32;
        private static final    int MAX_LENGTH  = 48;

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a chunk writer that will generate and write random chunk data.
         */

        private RandomChunkWriter( )
        {
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Writer interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk writer before any writing is performed.  The writer is reset at the start of
         * each pass of the document writer.
         *
         * @param  pass  the index of the pass (0 = first pass, 1 = second pass) of the document writer that
         *               calls this method.
         * @return {@code false} on the first pass to indicate that the chunk should not be rewritten on the
         *         second pass.
         */

        public boolean reset( int pass )
        {
            return false;
        }

        //--------------------------------------------------------------

        /**
         * Returns {@code -1} to indicate that the chunk size is unknown when the chunk header is written.
         * The size field in the chunk header will be fixed up after the chunk is written.
         *
         * @return {@code -1} to indicate that the chunk size is unknown when the chunk header is written.
         */

        public long getLength( )
        {
            return -1;
        }

        //--------------------------------------------------------------

        /**
         * Writes the chunk data to the specified data output, which is a {@code java.io.RandomAccessFile}.
         *
         * @param  dataOutput  the data output to which the chunk data are to be written.
         * @throws IOException
         *           if an I/O error occurs.
         */

        public void write( DataOutput dataOutput )
            throws IOException
        {
            Random rand = new Random( );
            byte[] data = new byte[MIN_LENGTH + rand.nextInt( MAX_LENGTH - MIN_LENGTH )];
            for ( int i = 0; i < data.length; ++i )
                data[i] = (byte)(NumberUtilities.HEX_DIGITS_UPPER.charAt( rand.nextInt( 16 ) ));
            crc.update( data );
            dataOutput.write( data );
        }

        //--------------------------------------------------------------

    }

    //==================================================================


    // TEXT-CHUNK WRITER CLASS


    /**
     * This class implements a chunk writer for a text chunk.  It writes some specified text, encoded as a
     * UTF-8 sequence, to an NLF document.
     */

    private class TextChunkWriter
        implements Chunk.Writer
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a chunk writer that will write the specified text.
         */

        private TextChunkWriter( String text )
        {
            this.text = text;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Writer interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk writer before any writing is performed.  The writer is reset at the start of
         * each pass of the document writer.
         *
         * @param  pass  the index of the pass (0 = first pass, 1 = second pass) of the document writer that
         *               calls this method.
         * @return {@code false} on the first pass to indicate that the chunk should not be rewritten on the
         *         second pass.
         */

        public boolean reset( int pass )
        {
            return false;
        }

        //--------------------------------------------------------------

        /**
         * Returns the length (size) of the chunk in bytes.
         *
         * @return the length (size) of the chunk in bytes.
         */

        public long getLength( )
        {
            return NlfUtilities.getUtf8Length( text );
        }

        //--------------------------------------------------------------

        /**
         * Writes the chunk data to the specified data output, which is a {@code java.io.RandomAccessFile}.
         *
         * @param  dataOutput  the data output to which the chunk data are to be written.
         * @throws IOException
         *           if an I/O error occurs.
         */

        public void write( DataOutput dataOutput )
            throws IOException
        {
            byte[] data = NlfUtilities.stringToUtf8( text );
            crc.update( data );
            dataOutput.write( data );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  text;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    /**
     * Constructs an instance of a {@code NlfProcessor} for the specified file.
     *
     * @param file  the Nested List File that will be generated and written by the {@link #write() write()}
     *              method or read and parsed by the {@link #read(java.io.File, java.io.File)
     *              read(java.io.File, java.io.File)} method.
     */

    public NlfProcessor( File file )
    {
        this.file = file;
        lineSeparator = SystemUtilities.getLineSeparator( );
        crc = new CRC32( );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    /**
     * Creates a string that will be used to indent a line of the text file that represents the structure
     * and content of an NLF.
     *
     * @param  indent    the indent index (ie, the number of levels of indentation, not the number of indent
     *                   characters).
     * @param  lineKind  the kind of line that is used in the representation of the tree structure of the
     *                   NLF: one of { LineKind.NONE, LineKind.VERTICAL, LineKind.HORIZONTAL }.
     * @return the indent string that was created from the specified indent and kind of line.
     */

    private static String createIndentString( int      indent,
                                              LineKind lineKind )
    {
        StringBuilder buffer = new StringBuilder( indent * INDENT_UNIT_WIDTH );
        for ( int i = 0; i < indent - 1; ++i )
        {
            buffer.append( '|' );
            for ( int j = 0; j < INDENT_UNIT_WIDTH - 1; ++j )
                buffer.append( ' ' );
        }

        if ( indent > 0 )
        {
            char char1 = ' ';
            char char2 = ' ';
            switch ( lineKind )
            {
                case VERTICAL:
                    char1 = '|';
                    break;

                case HORIZONTAL:
                    char1 = '+';
                    char2 = '-';
                    break;
            }
            buffer.append( char1 );
            for ( int i = 0; i < INDENT_UNIT_WIDTH - 2; ++i )
                buffer.append( char2 );
            buffer.append( ' ' );
        }

        return buffer.toString( );
    }

    //------------------------------------------------------------------

    /**
     * Converts an {@code NlfException} to an {@code AppException} and throws the resulting exception.
     *
     * @param  exception  the {@code NlfException} that is to be converted to an {@code AppException}.
     * @throws AppException
     *           derived from {@code exception}.
     */

    private static void throwAppException( NlfException exception )
        throws AppException
    {
        // Throw any underlying AppException
        Throwable cause = exception.getCause( );
        while ( cause != null )
        {
            if ( cause instanceof AppException )
                throw (AppException)cause;
            cause = cause.getCause( );
        }

        // Get file offset
        StringBuilder buffer = new StringBuilder( );
        long offset = exception.getOffset( );
        if ( offset >= 0 )
        {
            buffer.append( OFFSET_STR );
            buffer.append( Long.toHexString( offset ).toUpperCase( ) );
        }

        // Get exception detail message
        String message = exception.getMessage( );
        if ( message != null )
        {
            if ( buffer.length( ) > 0 )
                buffer.append( '\n' );
            buffer.append( message );
        }

        // Construct and throw exception
        AppException outException = new AppException( buffer.toString( ), exception.getCause( ) );
        throw (exception.getFile( ) == null) ? outException
                                             : new FileException( outException, exception.getFile( ) );
    }

    //------------------------------------------------------------------

    /**
     * Returns the kind of chunk that succeeds the specified chunk in its parent list.
     *
     * @param  chunk  the chunk whose successor is to be determined.
     * @return the kind of chunk that succeeds the specified chunk in its parent list: one of
     *         { ChunkKind.NONE, ChunkKind.CHUNK, ChunkKind.LIST }.
     */

    private static ChunkKind getNextChunkKind( Chunk chunk )
    {
        ChunkKind chunkKind = ChunkKind.NONE;
        ChunkList list = chunk.getParent( );
        if ( list != null )
        {
            for ( int i = 0; i < list.getNumChunks( ); ++i )
            {
                if ( list.getChunk( i ) == chunk )
                {
                    if ( i + 1 < list.getNumChunks( ) )
                        chunkKind = list.getChunk( i + 1 ).isList( ) ? ChunkKind.LIST : ChunkKind.CHUNK;
                    break;
                }
            }
        }
        return chunkKind;
    }

    //------------------------------------------------------------------

    /**
     * Returns whether a chunk contains text or binary data according to the "text" attribute of the
     * chunk's parent.
     *
     * @param  chunk  the chunk whose content is to be determined.
     * @return {@code true} if the chunk contains text; {@code false} otherwise.
     */

    private static boolean containsText( Chunk chunk )
    {
        boolean text = false;
        Attributes attrs = chunk.getParent( ).getAttributes( );
        if ( attrs != null )
        {
            int index = attrs.indexOf( AttrName.TEXT );
            if ( index >= 0 )
                text = Boolean.parseBoolean( attrs.getAttribute( index ).getValue( ) );
        }
        return text;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods : XmlFile.ElementWriter interface
////////////////////////////////////////////////////////////////////////

    /**
     * Writes an XML element using the specified {@code XmlWriter}.
     *
     * @param  writer   the {@code XmlWriter} that is to be used to write the element.
     * @param  element  the XML element that is to be written.
     * @param  indent   the number of characters by which the element is to be indented.
     * @throws IOException
     *           if an I/O error occurs.
     */

    public void writeElement( XmlWriter writer,
                              Element   element,
                              int       indent )
        throws IOException
    {
        final   int INDENT_INCREMENT    = 2;

        // Create list of attributes
        List<Attribute> attributes = new ArrayList<Attribute>( );
        NamedNodeMap attrs = element.getAttributes( );
        for ( int i = 0; i < attrs.getLength( ); ++i )
        {
            Node node = attrs.item( i );
            Attribute attribute = new Attribute( node.getNodeName( ), node.getNodeValue( ), true );
            if ( node.getNodeName( ).startsWith( "xmlns" ) )
                attributes.add( 0, attribute );
            else
                attributes.add( attribute );
        }

        // Write element that has character data
        NodeList nodes = element.getChildNodes( );
        if ( XmlUtilities.hasChildren( element, Node.TEXT_NODE ) )
        {
            // Write element start tag
            writer.writeElementStart( element.getTagName( ), attributes, indent, false, false );

            // Write text nodes
            for ( int i = 0; i < nodes.getLength( ); ++i )
            {
                Node node = nodes.item( i );
                if ( node.getNodeType( ) == Node.TEXT_NODE )
                    writer.writeEscaped( node.getNodeValue( ) );
            }

            // Write element end tag
            writer.writeElementEnd( element.getTagName( ), 0 );
        }

        // Write element that has no character data
        else
        {
            // Write element start tag
            writer.writeElementStart( element.getTagName( ), attributes, indent, true, false );

            // Write child elements
            for ( int i = 0; i < nodes.getLength( ); ++i )
            {
                Node node = nodes.item( i );
                if ( node.getNodeType( ) == Node.ELEMENT_NODE )
                    writeElement( writer, (Element)node, indent + INDENT_INCREMENT );
            }

            // Write element end tag
            writer.writeElementEnd( element.getTagName( ), indent );
        }
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    /**
     * Generates and writes an NLF document.  The document consists of eleven chunks on four levels: nine
     * chunks contain short pieces of text, which are encoded as UTF-8 sequences in the NLF; one chunk
     * contains a short random sequence of hexadecimal digit characters; one chunk contains the CRC value of
     * the data in the other chunks.
     * <p>
     * The document is intended to illustrate aspects of the {@code uk.org.blankaspect.nlf} rather than to
     * exemplify good design.  The CRC chunk appears at the start of the document in order to demonstrate
     * the two-pass method employed by the {@code uk.org.blankaspect.nlf.Document.write(java.io.File)}
     * method to write an NLF document.  The lists containing text chunks are nested for no reason other
     * than to show that lists can be nested.
     * </p>
     * <p>
     * The content of each chunk is written by a chunk writer that is appropriate for the content, which is
     * created and set on each chunk immediately after the chunk is created.
     * </p>
     *
     * @throws AppException
     *           if an error occurs in generating or writing the file.
     */

    public void write( )
        throws AppException
    {
        try
        {
            // Create document
            Document document = new Document( false );

            // Initialise CRC
            crc.reset( );

            // Create and populate root list
            ChunkList rootList = document.createRootList( ROOT_LIST_ID );
            rootList.setNamespaceName( NS_NAME_EXAMPLE_STR );

            Chunk chunk = document.createChunk( CRC_ID );
            chunk.setWriter( new CrcChunkWriter( ) );
            rootList.appendChunk( chunk );

            ChunkList alphaList = document.createList( ALPHA_LIST_ID ); // alpha list ...
            alphaList.setNamespaceName( NS_NAME_EUROPE_STR );
            rootList.appendChunk( alphaList );                          // ... is child of root list

            ChunkList deltaList = document.createList( DELTA_LIST_ID ); // delta list ...
            rootList.appendChunk( deltaList );                          // ... is child of root list

            // Add chunks to alpha list
            Attributes attrs = alphaList.createAttributes( );
            attrs.setAttribute( new Attributes.Attr( AttrName.TEXT, Boolean.TRUE.toString( ) ) );
            attrs.setAttribute( new Attributes.Attr( AttrName.LANGUAGE, ENGLISH_UK_STR ) );

            chunk = document.createChunk( BLAKE_ROSE_ID );
            chunk.setWriter( new TextChunkWriter( BLAKE_ROSE_STR ) );
            alphaList.appendChunk( chunk );

            chunk = document.createChunk( BLAKE_TYGER_ID );
            chunk.setWriter( new TextChunkWriter( BLAKE_TYGER_STR ) );
            alphaList.appendChunk( chunk );

            ChunkList betaList = document.createList( BETA_LIST_ID );   // beta list ...
            alphaList.appendChunk( betaList );                          // ... is child of alpha list

            chunk = document.createChunk( BLAKE_LONDON_ID );
            chunk.setWriter( new TextChunkWriter( BLAKE_LONDON_STR ) );
            alphaList.appendChunk( chunk );

            // Add chunks to beta list
            attrs = betaList.createAttributes( );
            attrs.setAttribute( new Attributes.Attr( AttrName.TEXT, Boolean.TRUE.toString( ) ) );
            attrs.setAttribute( new Attributes.Attr( AttrName.LANGUAGE, FRENCH_STR ) );

            chunk = document.createChunk( BAUDELAIRE_SERPENT_ID );
            chunk.setWriter( new TextChunkWriter( BAUDELAIRE_SERPENT_STR ) );
            betaList.appendChunk( chunk );

            ChunkList gammaList = document.createList( GAMMA_LIST_ID ); // gamma list ...
            gammaList.setNamespaceName( NS_NAME_USA_STR );
            betaList.appendChunk( gammaList );                          // ... is child of beta list

            chunk = document.createChunk( BAUDELAIRE_REVENANT_ID );
            chunk.setWriter( new TextChunkWriter( BAUDELAIRE_REVENANT_STR ) );
            betaList.appendChunk( chunk );

            chunk = document.createChunk( BAUDELAIRE_VOYAGE_ID );
            chunk.setWriter( new TextChunkWriter( BAUDELAIRE_VOYAGE_STR ) );
            betaList.appendChunk( chunk );

            // Add chunks to gamma list
            attrs = gammaList.createAttributes( );
            attrs.setAttribute( new Attributes.Attr( AttrName.TEXT, Boolean.TRUE.toString( ) ) );
            attrs.setAttribute( new Attributes.Attr( AttrName.LANGUAGE, ENGLISH_USA_STR ) );

            chunk = document.createChunk( WHITMAN_MYSELF_ID );
            chunk.setWriter( new TextChunkWriter( WHITMAN_MYSELF_STR ) );
            gammaList.appendChunk( chunk );

            chunk = document.createChunk( WHITMAN_FACES_ID );
            chunk.setWriter( new TextChunkWriter( WHITMAN_FACES_STR ) );
            gammaList.appendChunk( chunk );

            chunk = document.createChunk( WHITMAN_OSCEOLA_ID );
            chunk.setWriter( new TextChunkWriter( WHITMAN_OSCEOLA_STR ) );
            gammaList.appendChunk( chunk );

            // Add chunks to delta list
            attrs = deltaList.createAttributes( );
            attrs.setAttribute( new Attributes.Attr( AttrName.TEXT, Boolean.TRUE.toString( ) ) );

            chunk = document.createChunk( RANDOM_ID );
            chunk.setWriter( new RandomChunkWriter( ) );
            deltaList.appendChunk( chunk );

            // Update all chunk sizes
            rootList.updateSize( );

            // Write document
            document.write( file );
        }
        catch ( NlfException e )
        {
            throwAppException( e );     // convert NlfException to AppException
        }
    }

    //------------------------------------------------------------------

    /**
     * Reads and parses an NLF document, then converts it into a plain-text representation, which is written
     * to the specified output file.  If an XML file is specified, the NLF document is converted to XML,
     * and the resulting XML document is written to the the specified file.
     * <p>
     * This methods illustrates how an NLF document is read and parsed:
     * </p>
     * <ul>
     *   <li>the document is read and its structure is parsed;</li>
     *   <li>the chunk data are read using the default chunk reader that was set on each chunk when the
     *       document was parsed;</li>
     *   <li>if an XML document is generated, the chunk data are read again during the conversion to
     *       XML;</li>
     *   <li>the document is closed.</li>
     * </ul>
     *
     * @param  outFile  the file to which the plain-text representation of the NLF document is to be
     *                  written.  If this is {@code null}, the output is written to {@code System.out}.
     * @param  xmlFile  the file to which the XML document that results from the conversion of the NLF
     *                  document is to be written.  If this is {@code null}, no XML document is generated.
     * @throws AppException
     *           if an error occurs in reading or parsing the input file, or in writing either the output
     *           file or the XML file.
     */

    public void read( File outFile,
                      File xmlFile )
        throws AppException
    {
        // Initialise instance variables
        this.outFile = outFile;

        // Read input file and parse its structure
        Document document = null;
        try
        {
            document = new Document( false );
            document.read( file );
        }
        catch ( NlfException e )
        {
            throwAppException( e );     // convert NlfException to AppException
        }

        // Initialise CRC
        crc.reset( );

        // Read chunk data
        try
        {
            // If no output file, open output stream on standard output ...
            if ( outFile == null )
                outStream = new FileOutputStream( FileDescriptor.out );

            // ... otherwise, open output stream on file
            else
            {
                try
                {
                    outStream = new FileOutputStream( outFile );
                }
                catch ( FileNotFoundException e )
                {
                    throw new FileException( ErrorId.FAILED_TO_OPEN_FILE, outFile, e );
                }
                catch ( SecurityException e )
                {
                    throw new FileException( ErrorId.FILE_ACCESS_NOT_PERMITTED, outFile, e );
                }
            }

            // Read and print chunk data
            try
            {
                readList( document.getRootList( ), 0 );
            }
            catch ( IOException e )
            {
                throw new FileException( ErrorId.ERROR_READING_FILE, file, e );
            }

            // Read CRC chunk and test CRC value
            try
            {
                int index = document.getRootList( ).firstIndexOf( CRC_ID );
                if ( index < 0 )
                    throw new FileException( ErrorId.NO_CRC_CHUNK, file );
                Chunk crcChunk = document.getRootList( ).getChunk( index );
                crcChunk.getReader( ).reset( );
                long crcValue = crcChunk.getReader( ).getDataInput( ).readInt( ) & 0xFFFFFFFFL;
                if ( crcValue != crc.getValue( ) )
                    throw new FileException( ErrorId.INCORRECT_CRC, file );
                writeLine( CRC_CORRECT_STR );
            }
            catch ( IOException e )
            {
                throw new FileException( ErrorId.ERROR_READING_FILE, file, e );
            }

            // Generate and write XML file
            if ( xmlFile != null )
            {
                // Set encoder on each text chunk to encode chunk data as XML text node
                document.processChunks( NlfConstants.TraversalOrder.BREADTH_FIRST, new ChunkProcessor( ) );

                // Generate and write XML file
                try
                {
                    XmlFile.write( xmlFile, document.toXml( ), this );
                }
                catch ( NlfException e )
                {
                    throwAppException( e );     // convert NlfException to AppException
                }
            }

            // Close input file
            try
            {
                document.close( );
            }
            catch ( IOException e )
            {
                throw new FileException( ErrorId.FAILED_TO_CLOSE_FILE, file, e );
            }

            // Close output file
            if ( (outStream != null) && (outFile != null) )
            {
                try
                {
                    outStream.close( );
                    outStream = null;
                }
                catch ( IOException e )
                {
                    throw new FileException( ErrorId.FAILED_TO_CLOSE_FILE, outFile, e );
                }
            }
        }
        catch ( AppException e )
        {
            // Close input file
            if ( document != null )
                document.closeIgnoreException( );

            // Close output file
            if ( (outStream != null) && (outFile != null) )
            {
                try
                {
                    outStream.close( );
                }
                catch ( IOException e1 )
                {
                    // ignore
                }
            }

            // Rethrow exception
            throw e;
        }
    }

    //------------------------------------------------------------------

    /**
     * Reads and parses an NLF list, then writes its plain-text representation to the output stream.  This
     * method writes the child chunks of the list by calling {@link #readChunk(uk.org.blankaspect.nlf.Chunk,
     * int) readChunk(uk.org.blankaspect.nlf.Chunk, int)} on each general chunk, and calling itself on each
     * sublist.
     *
     * @param  list    the NLF list that is to be written.
     * @param  indent  the indent index (ie, the number of levels of indentation, not the number of indent
     *                 characters).
     * @throws AppException
     *           if an error occurs in parsing the input file or in writing to the output stream.
     * @throws IOException
     *           if an error occurs in reading the input file.
     */

    private void readList( ChunkList list,
                           int       indent )
        throws AppException, IOException
    {
        // Write list info
        String indentStr = createIndentString( indent,
                                               list.hasChunks( ) ? LineKind.VERTICAL : LineKind.NONE );
        writeLine( createIndentString( indent, LineKind.HORIZONTAL ) + LIST_INSTANCE_ID_STR +
                                                                    EQUALS_STR + list.getInstanceId( ) );
        String namespaceName = list.getLocalNamespaceName( );
        if ( namespaceName == null )
            namespaceName = new String( );
        writeLine( indentStr + LOCAL_NAMESPACE_NAME_STR + EQUALS_STR + namespaceName );
        writeLine( indentStr + NAME_STR + EQUALS_STR + list.getName( ) );
        writeLine( indentStr + PATHNAME_STR + EQUALS_STR + list.getPathname( ) );

        // Write attributes
        Attributes attrs = list.getAttributes( );
        if ( attrs != null )
        {
            for ( int i = 0; i < attrs.getNumAttributes( ); ++i )
            {
                Attributes.Attr attr = attrs.getAttribute( i );
                writeLine( indentStr + ATTR_STR + " " + attr.getName( ) + EQUALS_STR  + "\"" +
                                                                                attr.getValue( ) + "\"" );
            }
        }

        // Process children of list
        if ( list.hasChunks( ) )
        {
            // Write empty line
            writeLine( createIndentString( indent, LineKind.VERTICAL ) + "|" );

            // Process children
            for ( int i = 0; i < list.getNumChunks( ); ++i )
            {
                Chunk chunk = list.getChunk( i );
                if ( chunk.isList( ) )
                    readList( (ChunkList)chunk, indent + 1 );
                else
                {
                    if ( !chunk.isAttributes( ) )
                        readChunk( chunk, indent + 1 );
                }
            }
        }

        // Write separator
        if ( list.isRoot( ) || (getNextChunkKind( list ) != ChunkKind.NONE) )
        {
            int length = SEPARATOR_LENGTH - indent * INDENT_UNIT_WIDTH;
            writeLine( indentStr + new String( StringUtilities.createCharArray( '=', length ) ) );
        }
    }

    //------------------------------------------------------------------

    /**
     * Reads and parses an NLF chunk, then writes its plain-text representation to the output stream.
     *
     * @param  chunk   the NLF chunk that is to be written.
     * @param  indent  the indent index (ie, the number of levels of indentation, not the number of indent
     *                 characters).
     * @throws AppException
     *           if an error occurs in parsing the input file or in writing to the output stream.
     * @throws IOException
     *           if an error occurs in reading the input file.
     */

    private void readChunk( Chunk chunk,
                            int   indent )
        throws AppException, IOException
    {
        // Get kind of next chunk in parent list
        ChunkKind nextChunkKind = getNextChunkKind( chunk );

        // Write chunk info
        String indentStr = createIndentString( indent,
                                               (nextChunkKind == ChunkKind.NONE) ? LineKind.NONE
                                                                                 : LineKind.VERTICAL );

        writeLine( createIndentString( indent, LineKind.HORIZONTAL ) + CHUNK_ID_STR + EQUALS_STR +
                                                                                        chunk.getId( ) );
        writeLine( indentStr + NAME_STR + EQUALS_STR + chunk.getName( ) );
        writeLine( indentStr + PATHNAME_STR + EQUALS_STR + chunk.getPathname( ) );
        writeLine( indentStr + SIZE_STR + EQUALS_STR + chunk.getSize( ) + " (0x" +
                                                Long.toHexString( chunk.getSize( ) ).toUpperCase( ) + ")" );

        // Reset chunk reader
        chunk.getReader( ).reset( );

        // Read chunk data
        if ( chunk.getSize( ) > Integer.MAX_VALUE )
            throw new FileException( ErrorId.CHUNK_DATA_ARE_TOO_LARGE, file );
        byte[] buffer = new byte[(int)chunk.getSize( )];
        chunk.getReader( ).getDataInput( ).readFully( buffer );

        // Write chunk data
        write( indentStr + DATA_STR + EQUALS_STR );
        if ( containsText( chunk ) )
        {
            try
            {
                writeLine( "\"" + NlfUtilities.utf8ToString( buffer ) + "\"" );
            }
            catch ( IllegalArgumentException e )
            {
                throw new FileException( ErrorId.INVALID_TEXT, file );
            }
            crc.update( buffer );
        }
        else
        {
            write( HEX_STR );
            int length = Math.min( buffer.length, MAX_BINARY_DATA_LENGTH );
            for ( int i = 0; i < length; ++i )
            {
                write( " " );
                write( NumberUtilities.byteToHexString( buffer[i] ) );
            }
            if ( length < buffer.length )
                write( " ..." );
            writeLine( );
        }

        // Write separator
        if ( nextChunkKind != ChunkKind.NONE )
        {
            char ch = (nextChunkKind == ChunkKind.CHUNK) ? '-' : '=';
            int length = SEPARATOR_LENGTH - indent * INDENT_UNIT_WIDTH;
            writeLine( indentStr + new String( StringUtilities.createCharArray( ch, length ) ) );
        }
    }

    //------------------------------------------------------------------

    /**
     * Writes the specified string to the output stream.
     *
     * @param  str  the string that is to be written.
     * @throws AppException
     *           if an error occurs in writing to the output stream.
     */

    private void write( String str )
        throws AppException
    {
        try
        {
            outStream.write( (outFile == null) ? str.getBytes( ) : str.getBytes( ENCODING_NAME_UTF8 ) );
        }
        catch ( IOException e )
        {
            throw new FileException( ErrorId.ERROR_WRITING_FILE, outFile, e );
        }
    }

    //------------------------------------------------------------------

    /**
     * Writes a line separator to the output stream.
     *
     * @throws AppException
     *           if an error occurs in writing to the output stream.
     */

    private void writeLine( )
        throws AppException
    {
        write( lineSeparator );
    }

    //------------------------------------------------------------------

    /**
     * Writes the specified string followed by a line separator to the output stream.
     *
     * @param  str  the string that is to be written.
     * @throws AppException
     *           if an error occurs in writing to the output stream.
     */

    private void writeLine( String str )
        throws AppException
    {
        write( str );
        writeLine( );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    private File                file;
    private String              lineSeparator;
    private CRC32               crc;
    private File                outFile;
    private FileOutputStream    outStream;

}

//----------------------------------------------------------------------
