/*====================================================================*\

PropertyString.java

Property string class.

------------------------------------------------------------------------

This file is part of NlfExample, a Java application that demonstrates
the use of the uk.org.blankaspect.nlf API.
Copyright 2007-2013 Andy Morgan-Richards.

NlfExample is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package util;

//----------------------------------------------------------------------


// IMPORTS


import java.io.File;

//----------------------------------------------------------------------


// PROPERTY STRING CLASS


public class PropertyString
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    public static final     String  KEY_PREFIX  = "${";
    public static final     String  KEY_SUFFIX  = "}";

    public static final     char    KEY_SEPARATOR_CHAR  = '.';
    public static final     String  KEY_SEPARATOR       = Character.toString( KEY_SEPARATOR_CHAR );

    public static final     String  USER_HOME_PREFIX    = "~";

    private static final    int KEY_PREFIX_LENGTH   = KEY_PREFIX.length( );
    private static final    int KEY_SUFFIX_LENGTH   = KEY_SUFFIX.length( );

    private static final    String  ENV_PREFIX_KEY  = "app" + KEY_SEPARATOR + "envPrefix";
    private static final    String  SYS_PREFIX_KEY  = "app" + KEY_SEPARATOR + "sysPrefix";

    private static final    String  DEFAULT_ENV_PREFIX          = "env" + KEY_SEPARATOR;
    private static final    String  DEFAULT_SYS_PREFIX          = "sys" + KEY_SEPARATOR;
    private static final    String  DEFAULT_CONFIGURATION_NAME  = "blankaspect";

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    private PropertyString( )
    {
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    public static String getEnvPrefix( )
    {
        try
        {
            String value = System.getProperty( ENV_PREFIX_KEY );
            if ( value != null )
                envPrefix = value.endsWith( KEY_SEPARATOR ) ? value : value + KEY_SEPARATOR;
        }
        catch ( SecurityException e )
        {
            // ignore
        }
        return envPrefix;
    }

    //------------------------------------------------------------------

    public static String getSysPrefix( )
    {
        try
        {
            String value = System.getProperty( SYS_PREFIX_KEY );
            if ( value != null )
                sysPrefix = value.endsWith( KEY_SEPARATOR ) ? value : value + KEY_SEPARATOR;
        }
        catch ( SecurityException e )
        {
            // ignore
        }
        return sysPrefix;
    }

    //------------------------------------------------------------------

    public static String getConfigurationName( )
    {
        Package pkg = PropertyString.class.getPackage( );
        if ( pkg != null )
        {
            String[] strs = pkg.getName( ).split( "." );
            if ( strs.length > 1 )
                configurationName = strs[1];
        }
        return configurationName;
    }

    //------------------------------------------------------------------

    public static String concatenateKeys( CharSequence... keys )
    {
        // Calculate length of buffer
        int length = -1;
        for ( CharSequence key : keys )
            length += key.length( ) + 1;

        // Concatenate keys
        StringBuilder buffer = new StringBuilder( length );
        for ( int i = 0; i < keys.length; ++i )
        {
            if ( i > 0 )
                buffer.append( KEY_SEPARATOR_CHAR );
            buffer.append( keys[i] );
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String parse( String str )
    {
        // Test for null string
        if ( str == null )
            return null;

        // Get environment and system prefixes
        getEnvPrefix( );
        getSysPrefix( );

        // Parse string
        StringBuilder buffer = new StringBuilder( str );
        int index = 0;
        while ( index < buffer.length( ) )
        {
            index = buffer.indexOf( KEY_PREFIX, index );
            if ( index < 0 )
                break;
            index += KEY_PREFIX_LENGTH;
            int startIndex = index;
            index = buffer.indexOf( KEY_SUFFIX, index );
            if ( index < 0 )
                index = startIndex;
            else
            {
                String propertyValue = new String( );
                if ( index > startIndex )
                {
                    String key = buffer.substring( startIndex, index );
                    try
                    {
                        if ( key.startsWith( envPrefix ) )
                        {
                            String value = System.getenv( key.substring( envPrefix.length( ) ) );
                            if ( value != null )
                                propertyValue = value;
                        }
                        else if ( key.startsWith( sysPrefix ) )
                        {
                            String value = System.getProperty( key.substring( sysPrefix.length( ) ) );
                            if ( value != null )
                                propertyValue = value;
                        }
                        else
                        {
                            String value = System.getProperty( key );
                            if ( value == null )
                                value = System.getenv( key );
                            if ( value != null )
                                propertyValue = value;
                        }
                    }
                    catch ( SecurityException e )
                    {
                        // ignore
                    }
                }
                buffer.replace( startIndex - KEY_PREFIX_LENGTH, index + KEY_SUFFIX_LENGTH, propertyValue );
                index = startIndex - KEY_PREFIX_LENGTH + propertyValue.length( );
            }
        }

        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String parsePathname( String str )
    {
        if ( str.startsWith( USER_HOME_PREFIX ) )
        {
            int prefixLength = USER_HOME_PREFIX.length( );
            if ( (str.length( ) == prefixLength) || (str.charAt( prefixLength ) == File.separatorChar) ||
                 (str.charAt( prefixLength ) == '/') )
            try
            {
                String pathname = SystemUtilities.getUserHomePathname( );
                if ( pathname != null )
                    str = pathname + str.substring( prefixLength );
            }
            catch ( SecurityException e )
            {
                // ignore
            }
        }
        return parse( str );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class variables
////////////////////////////////////////////////////////////////////////

    private static  String  envPrefix           = DEFAULT_ENV_PREFIX;
    private static  String  sysPrefix           = DEFAULT_SYS_PREFIX;
    private static  String  configurationName   = DEFAULT_CONFIGURATION_NAME;

}

//----------------------------------------------------------------------
