/*====================================================================*\

StringUtilities.java

String utility methods class.

------------------------------------------------------------------------

This file is part of NlfExample, a Java application that demonstrates
the use of the uk.org.blankaspect.nlf API.
Copyright 2007-2013 Andy Morgan-Richards.

NlfExample is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package util;

//----------------------------------------------------------------------


// IMPORTS


import java.io.UnsupportedEncodingException;

import java.nio.ByteBuffer;

import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.UnsupportedCharsetException;

import java.util.Arrays;
import java.util.List;

//----------------------------------------------------------------------


// STRING UTILITY METHODS CLASS


public class StringUtilities
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    public static final     String  ENCODING_NAME_UTF8              = "UTF-8";
    public static final     String  SUBSTITUTION_PLACEHOLDER_PREFIX = "%";

    private static final    int MIN_SUBSTITUTION_INDEX  = 1;

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    private StringUtilities( )
    {
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    public static boolean isNullOrEmpty( String str )
    {
        return ( (str == null) || str.isEmpty( ) );
    }

    //------------------------------------------------------------------

    public static boolean equal( String str1,
                                 String str2 )
    {
        return ( (str1 == null) ? (str2 == null) : str1.equals( str2 ) );
    }

    //------------------------------------------------------------------

    public static int getIndex( String[] strs,
                                String   str )
    {
        for ( int i = 0; i < strs.length; ++i )
        {
            if ( strs[i].equals( str ) )
                return i;
        }
        return -1;
    }

    //------------------------------------------------------------------

    public static char[] createCharArray( char ch,
                                          int  length )
    {
        char[] array = new char[length];
        Arrays.fill( array, ch );
        return array;
    }

    //------------------------------------------------------------------

    public static String createCharString( char ch,
                                           int  length )
    {
        return new String( createCharArray( ch, length ) );
    }

    //------------------------------------------------------------------

    public static String firstCharToLowerCase( String str )
    {
        return ( str.substring( 0, 1 ).toLowerCase( ) + str.substring( 1 ) );
    }

    //------------------------------------------------------------------

    public static String firstCharsToLowerCase( String str )
    {
        StringBuilder buffer = new StringBuilder( str );
        boolean start = true;
        for ( int i = 0; i < buffer.length( ); ++i )
        {
            char ch = buffer.charAt( i );
            if ( start )
                buffer.setCharAt( i, Character.toLowerCase( ch ) );
            start = (ch == ' ') || (ch == '-');
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String firstCharToUpperCase( String str )
    {
        return ( str.substring( 0, 1 ).toUpperCase( ) + str.substring( 1 ) );
    }

    //------------------------------------------------------------------

    public static String firstCharsToUpperCase( String str )
    {
        StringBuilder buffer = new StringBuilder( str );
        boolean start = true;
        for ( int i = 0; i < buffer.length( ); ++i )
        {
            char ch = buffer.charAt( i );
            if ( start )
                buffer.setCharAt( i, Character.toUpperCase( ch ) );
            start = (ch == ' ') || (ch == '-');
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static int getMaximumLength( List<? extends CharSequence> seqs )
    {
        int maxLength = 0;
        for ( CharSequence seq : seqs )
        {
            int length = seq.length( );
            if ( maxLength < length )
                maxLength = length;
        }
        return maxLength;
    }

    //------------------------------------------------------------------

    public static int getMaximumLength( String[] strs )
    {
        return getMaximumLength( Arrays.asList( strs ) );
    }

    //------------------------------------------------------------------

    public static String stripBefore( CharSequence seq )
    {
        int index = 0;
        while ( index < seq.length( ) )
        {
            if ( !Character.isWhitespace( seq.charAt( index ) ) )
                break;
            ++index;
        }
        return seq.subSequence( index, seq.length( ) ).toString( );
    }

    //------------------------------------------------------------------

    public static String stripAfter( CharSequence seq )
    {
        int index = seq.length( ) - 1;
        while ( index >= 0 )
        {
            if ( !Character.isWhitespace( seq.charAt( index ) ) )
                break;
            --index;
        }
        return seq.subSequence( 0, index + 1 ).toString( );
    }

    //------------------------------------------------------------------

    public static String padBefore( CharSequence seq,
                                    int          length )
    {
        return padBefore( seq, length, ' ' );
    }

    //------------------------------------------------------------------

    public static String padBefore( CharSequence seq,
                                    int          length,
                                    char         ch )
    {
        int padLength = length - seq.length( );
        return ( (padLength > 0) ? new String( createCharArray( ch, padLength ) ) + seq.toString( )
                                 : seq.toString( ) );
    }

    //------------------------------------------------------------------

    public static String padAfter( CharSequence seq,
                                   int          length )
    {
        return padAfter( seq, length, ' ' );
    }

    //------------------------------------------------------------------

    public static String padAfter( CharSequence seq,
                                   int          length,
                                   char         ch )
    {
        int padLength = length - seq.length( );
        return ( (padLength > 0) ? seq.toString( ) + new String( createCharArray( ch, padLength ) )
                                 : seq.toString( ) );
    }

    //------------------------------------------------------------------

    public static String concatenate( CharSequence                 separator,
                                      List<? extends CharSequence> seqs )
    {
        // Calculate length of buffer
        int length = (separator == null) ? 0 : separator.length( ) * seqs.size( );
        if ( length > 0 )
            --length;
        for ( CharSequence seq : seqs )
            length += seq.length( );

        // Concatenate character sequences
        StringBuilder buffer = new StringBuilder( length );
        for ( int i = 0; i < seqs.size( ); ++i )
        {
            if ( (separator != null) && (i > 0) )
                buffer.append( separator );
            buffer.append( seqs.get( i ) );
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String concatenate( CharSequence separator,
                                      String...    strs )
    {
        return concatenate( separator, Arrays.asList( strs ) );
    }

    //------------------------------------------------------------------

    public static String substitute( String    str,
                                     String... substitutionStrs )
    {
        StringBuilder buffer = new StringBuilder( str );
        int index = 0;
        while ( index < buffer.length( ) )
        {
            index = buffer.indexOf( SUBSTITUTION_PLACEHOLDER_PREFIX, index );
            if ( index < 0 )
                index = buffer.length( );
            else
                ++index;
            int startIndex = index;
            while ( index < buffer.length( ) )
            {
                char ch = buffer.charAt( index );
                if ( (ch < '0') || (ch > '9') )
                    break;
                ++index;
            }
            if ( index > startIndex )
            {
                int substIndex = Integer.parseInt( buffer.substring( startIndex, index ) ) -
                                                                                    MIN_SUBSTITUTION_INDEX;
                if ( (substIndex >= 0) && (substIndex < substitutionStrs.length) )
                {
                    String substStr = substitutionStrs[substIndex];
                    if ( substStr == null )
                        substStr = new String( );
                    --startIndex;
                    buffer.replace( startIndex, index, substStr );
                    index = startIndex + substStr.length( );
                }
            }
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String substituteInteger( String str,
                                            int    value )
    {
        return substitute( str, Integer.toString( value ) );
    }

    //------------------------------------------------------------------

    public static String applyPrefix( String prefix,
                                      String str )
    {
        if ( isNullOrEmpty( prefix ) )
            return str;
        return ( prefix + (prefix.endsWith( "_" ) ? str : firstCharToUpperCase( str )) );
    }

    //------------------------------------------------------------------

    /**
     * Decodes the specified UTF-8 sequence to a string.  An {@code IllegalArgumentException} is thrown if
     * the UTF-8 sequence is malformed or contains bytes that cannot be mapped to a character.  This method
     * is equivalent to {@code #utf8ToString( data, 0, data.length )}.
     *
     * @param  data  an array that contains the UTF-8 sequence that is to be decoded.
     * @return the string that results from decoding the input sequence.
     * @throws IllegalArgumentException
     *           if
     *           <ul>
     *             <li>the UTF-8 sequence is malformed, or</li>
     *             <li>the UTF-8 sequence contains bytes that cannot be mapped to a character.</li>
     *           </ul>
     * @throws UnsupportedCharsetException
     *           if the Java implementation does not support the UTF-8 character encoding, which is required
     *           of all Java implementations.
     */

    public static String utf8ToString( byte[] data )
        throws IllegalArgumentException, UnsupportedCharsetException
    {
        return utf8ToString( data, 0, data.length );
    }

    //------------------------------------------------------------------

    /**
     * Decodes the specified UTF-8 sequence to a string.  An {@code IllegalArgumentException} is thrown if
     * the UTF-8 sequence is malformed or if it contains bytes that cannot be mapped to a character.
     *
     * @param  data    an array that contains the UTF-8 sequence that is to be decoded.
     * @param  offset  the offset to {@code data} at which the input sequence begins.
     * @param  length  the length of the input sequence.
     * @return the string that results from decoding the input sequence.
     * @throws IllegalArgumentException
     *           if
     *           <ul>
     *             <li>the UTF-8 sequence is malformed, or</li>
     *             <li>the UTF-8 sequence contains bytes that cannot be mapped to a character.</li>
     *           </ul>
     * @throws UnsupportedCharsetException
     *           if the Java implementation does not support the UTF-8 character encoding, which is required
     *           of all Java implementations.
     */

    public static String utf8ToString( byte[] data,
                                       int    offset,
                                       int    length )
        throws IllegalArgumentException, UnsupportedCharsetException
    {
        try
        {
            CharsetDecoder decoder = Charset.forName( ENCODING_NAME_UTF8 ).newDecoder( ).
                                                        onMalformedInput( CodingErrorAction.REPORT ).
                                                        onUnmappableCharacter( CodingErrorAction.REPORT );
            return decoder.decode( ByteBuffer.wrap( data, offset, length ) ).toString( );
        }
        catch ( UnsupportedCharsetException e )
        {
            throw e;
        }
        catch ( Exception e )
        {
            throw new IllegalArgumentException( );
        }
    }

    //------------------------------------------------------------------

    /**
     * Encodes the specified string as a UTF-8 sequence.
     *
     * @param  str  the string that is to be encoded.
     * @return an array containing the UTF-8 sequence that results from encoding the input string.
     * @throws UnsupportedCharsetException
     *           if the Java implementation does not support the UTF-8 character encoding, which is required
     *           of all Java implementations.
     */

    public static byte[] stringToUtf8( String str )
        throws UnsupportedCharsetException
    {
        try
        {
            return str.getBytes( ENCODING_NAME_UTF8 );
        }
        catch ( UnsupportedEncodingException e )
        {
            throw new UnsupportedCharsetException( ENCODING_NAME_UTF8 );
        }
    }

    //------------------------------------------------------------------

    public static String substituteEnvironmentVariables( String str,
                                                         String prefix,
                                                         String suffix )
    {
        StringBuilder buffer = new StringBuilder( str );
        int prefixLength = prefix.length( );
        int index = 0;
        while ( index < buffer.length( ) )
        {
            index = buffer.indexOf( prefix, index );
            if ( index < 0 )
                break;
            index += prefixLength;
            int startIndex = index;
            index = buffer.indexOf( suffix, index );
            if ( index < 0 )
                index = startIndex;
            else
            {
                String value = null;
                if ( index > startIndex )
                {
                    try
                    {
                        value = System.getenv( buffer.substring( startIndex, index ) );
                    }
                    catch ( SecurityException e )
                    {
                        // ignore
                    }
                }
                startIndex -= prefixLength;
                if ( value == null )
                    value = new String( );
                buffer.replace( startIndex, index + suffix.length( ), value );
                index = startIndex + value.length( );
            }
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

}

//----------------------------------------------------------------------
