/*====================================================================*\

XmlUtilities.java

XML utilities class.

------------------------------------------------------------------------

This file is part of NlfExample, a Java application that demonstrates
the use of the uk.org.blankaspect.nlf API.
Copyright 2007-2013 Andy Morgan-Richards.

NlfExample is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package xml;

//----------------------------------------------------------------------


// IMPORTS


import exception.AppException;

import java.io.ByteArrayInputStream;
import java.io.CharArrayReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.StringReader;

import java.net.URI;

import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

//----------------------------------------------------------------------


// XML UTILITIES CLASS


public class XmlUtilities
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    private static final    String  XML_DECLARATION_PREFIX  = "<?xml";

    private interface AttrName
    {
        String  XMLNS   = "xmlns";
    }

////////////////////////////////////////////////////////////////////////
//  Enumerated types
////////////////////////////////////////////////////////////////////////


    // ERROR IDENTIFIERS


    private enum ErrorId
        implements AppException.Id
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        FAILED_TO_INSTANTIATE_DOCUMENT_BUILDER_FACTORY
        ( "Failed to instantiate a DOM document builder factory." ),

        FAILED_TO_CREATE_DOCUMENT_BUILDER
        ( "Failed to create a DOM document builder." ),

        ERROR_PARSING_FILE
        ( "An error occurred while parsing the file." );

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        private ErrorId( String message )
        {
            this.message = message;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : AppException.Id interface
    ////////////////////////////////////////////////////////////////////

        public String getMessage( )
        {
            return message;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  message;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    private XmlUtilities( )
    {
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    public static ErrorLogger getErrorHandler( )
    {
        if ( errorHandler == null )
            errorHandler = new ErrorLogger( );
        return errorHandler;
    }

    //------------------------------------------------------------------

    public static String escape( char ch )
    {
        switch ( ch )
        {
            case '<':
                return "&lt;";

            case '>':
                return "&gt;";

            case '\'':
                return "&apos;";

            case '"':
                return "&quot;";

            case '&':
                return "&amp;";

            default:
                return Character.toString( ch );
        }
    }

    //------------------------------------------------------------------

    public static String escape( CharSequence charSeq )
    {
        StringBuilder buffer = new StringBuilder( );
        for ( int i = 0; i < charSeq.length( ); ++i )
            buffer.append( escape( charSeq.charAt( i ) ) );
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String concatenatePath( String... strs )
    {
        StringBuilder buffer = new StringBuilder( 256 );
        for ( int i = 0; i < strs.length; ++i )
        {
            if ( i > 0 )
                buffer.append( XmlConstants.PATH_SEPARATOR_CHAR );
            buffer.append( strs[i] );
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static String getElementPath( Element element )
    {
        List<String> strs = new ArrayList<String>( );
        while ( element != null )
        {
            strs.add( element.getNodeName( ) );
            Node parent = element.getParentNode( );
            element = (parent instanceof Element) ? (Element)parent : null;
        }

        StringBuilder buffer = new StringBuilder( strs.size( ) << 4 );
        for ( int i = strs.size( ) - 1; i >= 0; --i )
        {
            buffer.append( strs.get( i ) );
            if ( i > 0 )
                buffer.append( XmlConstants.PATH_SEPARATOR_CHAR );
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

    public static boolean hasChildren( Element element,
                                       int     nodeType )
    {
        if ( element.hasChildNodes( ) )
        {
            NodeList nodes = element.getChildNodes( );
            for ( int i = 0; i < nodes.getLength( ); ++i )
            {
                if ( nodes.item( i ).getNodeType( ) == nodeType )
                    return true;
            }
        }
        return false;
    }

    //------------------------------------------------------------------

    public static boolean isXml( File file )
        throws FileNotFoundException, IOException, SecurityException
    {
        if ( file.length( ) < XML_DECLARATION_PREFIX.length( ) )
            return false;

        byte[] buffer = new byte[XML_DECLARATION_PREFIX.length( )];
        RandomAccessFile raFile = new RandomAccessFile( file, "r" );
        raFile.readFully( buffer );
        raFile.close( );
        return XML_DECLARATION_PREFIX.equals( new String( buffer, XmlConstants.ENCODING_NAME_UTF8 ) );
    }

    //------------------------------------------------------------------

    public static DocumentBuilder createDocumentBuilder( boolean validate )
        throws AppException
    {
        try
        {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance( );
            docBuilderFactory.setValidating( validate );
            docBuilderFactory.setNamespaceAware( true );
            docBuilderFactory.setXIncludeAware( true );
            return docBuilderFactory.newDocumentBuilder( );
        }
        catch ( FactoryConfigurationError e )
        {
            throw new AppException( ErrorId.FAILED_TO_INSTANTIATE_DOCUMENT_BUILDER_FACTORY, e );
        }
        catch ( ParserConfigurationException e )
        {
            throw new AppException( ErrorId.FAILED_TO_CREATE_DOCUMENT_BUILDER, e );
        }
    }

    //------------------------------------------------------------------

    public static Document createDocument( )
        throws AppException
    {
        return createDocumentBuilder( false ).newDocument( );
    }

    //------------------------------------------------------------------

    public static Document createDocument( byte[] data )
        throws AppException
    {
        return createDocument( data, null, false );
    }

    //------------------------------------------------------------------

    public static Document createDocument( byte[]  data,
                                           URI     baseUri,
                                           boolean validate )
        throws AppException
    {
        InputSource inputSource = new InputSource( new ByteArrayInputStream( data ) );
        inputSource.setEncoding( XmlConstants.ENCODING_NAME_UTF8 );
        return createDocument( inputSource, baseUri, validate );
    }

    //------------------------------------------------------------------

    public static Document createDocument( char[] data )
        throws AppException
    {
        return createDocument( data, null, false );
    }

    //------------------------------------------------------------------

    public static Document createDocument( char[]  data,
                                           URI     baseUri,
                                           boolean validate )
        throws AppException
    {
        return createDocument( new InputSource( new CharArrayReader( data ) ), baseUri, validate );
    }

    //------------------------------------------------------------------

    public static Document createDocument( String data )
        throws AppException
    {
        return createDocument( data, null, false );
    }

    //------------------------------------------------------------------

    public static Document createDocument( String  data,
                                           URI     baseUri,
                                           boolean validate )
        throws AppException
    {
        return createDocument( new InputSource( new StringReader( data ) ), baseUri, validate );
    }

    //------------------------------------------------------------------

    public static Document createDocument( InputStream inStream )
        throws AppException
    {
        return createDocument( inStream, null, false );
    }

    //------------------------------------------------------------------

    public static Document createDocument( InputStream inStream,
                                           URI         baseUri,
                                           boolean     validate )
        throws AppException
    {
        return createDocument( new InputSource( inStream ), baseUri, validate );
    }

    //------------------------------------------------------------------

    public static Document createDocument( InputSource inputSource,
                                           URI         baseUri,
                                           boolean     validate )
        throws AppException
    {
        try
        {
            if ( baseUri != null )
                inputSource.setSystemId( baseUri.toString( ) );
            DocumentBuilder documentBuilder = createDocumentBuilder( validate );
            getErrorHandler( ).clear( );
            documentBuilder.setErrorHandler( getErrorHandler( ) );
            return documentBuilder.parse( inputSource );
        }
        catch ( SAXException e )
        {
            throw new AppException( ErrorId.ERROR_PARSING_FILE, e );
        }
        catch ( IOException e )
        {
            throw new AppException( ErrorId.ERROR_PARSING_FILE, e );
        }
    }

    //------------------------------------------------------------------

    public static String getAttribute( Element element,
                                       String  attributeName )
    {
        return ( element.hasAttribute( attributeName ) ? element.getAttribute( attributeName ) : null );
    }

    //------------------------------------------------------------------

    public static String getAttribute( Element element,
                                       String  namespaceName,
                                       String  attributeName )
    {
        return ( element.hasAttributeNS( namespaceName, attributeName )
                                                    ? element.getAttributeNS( namespaceName, attributeName )
                                                    : null );
    }

    //------------------------------------------------------------------

    public static List<Element> getChildElements( Element element )
    {
        List<Element> elements = new ArrayList<Element>( );
        NodeList nodes = element.getChildNodes( );
        for ( int i = 0; i < nodes.getLength( ); ++i )
        {
            Node node = nodes.item( i );
            if ( node.getNodeType( ) == Node.ELEMENT_NODE )
                elements.add( (Element)node );
        }
        return elements;
    }

    //------------------------------------------------------------------

    public static Element findElement( Element element,
                                       String  elementPath )
    {
        return findElement( null, element, elementPath, false );
    }

    //------------------------------------------------------------------

    public static Element findElement( Document document,
                                       String   elementPath )
    {
        return findElement( document, elementPath, false );
    }

    //------------------------------------------------------------------

    public static Element findElement( Document document,
                                       String   elementPath,
                                       boolean  create )
    {
        return findElement( document, document.getDocumentElement( ), elementPath, create );
    }

    //------------------------------------------------------------------

    public static Element findElement( Document document,
                                       Element  element,
                                       String   elementPath,
                                       boolean  create )
    {
        Node node = element;
        String[] elementPathComponents = elementPath.split( "\\" + XmlConstants.PATH_SEPARATOR, -1 );
        int pathIndex = 0;
        while ( pathIndex < elementPathComponents.length )
        {
            String elementName = elementPathComponents[pathIndex];
            Node parentNode = node;
            NodeList nodes = node.getChildNodes( );
            int nodeIndex = 0;
            while ( nodeIndex < nodes.getLength( ) )
            {
                node = nodes.item( nodeIndex );
                if ( (node.getNodeType( ) == Node.ELEMENT_NODE) &&
                     node.getNodeName( ).equals( elementName ) )
                    break;
                ++nodeIndex;
            }
            if ( nodeIndex >= nodes.getLength( ) )
            {
                if ( !create )
                    break;
                parentNode.appendChild( document.createElement( elementName ) );
                node = parentNode.getLastChild( );
            }
            ++pathIndex;
        }

        return ( (pathIndex < elementPathComponents.length) ? null : (Element)node );
    }

    //------------------------------------------------------------------

    public static void setNamespaceAttribute( Element element )
    {
        // Set namespace declaration attribute on this element
        String namespaceName = element.getNamespaceURI( );
        if ( namespaceName != null )
        {
            Node parent = element.getParentNode( );
            if ( (parent == null) || !namespaceName.equals( parent.getNamespaceURI( ) ) )
            {
                String prefix = element.getPrefix( );
                String attrName = (prefix == null) ? AttrName.XMLNS : AttrName.XMLNS + ":" + prefix;
                element.setAttribute( attrName, namespaceName );
            }
        }

        // Set namespace declaration attribute on child elements
        NodeList nodes = element.getChildNodes( );
        for ( int i = 0; i < nodes.getLength( ); ++i )
        {
            Node node = nodes.item( i );
            if ( node.getNodeType( ) == Node.ELEMENT_NODE )
                setNamespaceAttribute( (Element)node );
        }
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class variables
////////////////////////////////////////////////////////////////////////

    private static  ErrorLogger errorHandler;

}

//----------------------------------------------------------------------
