/*====================================================================*\

XmlWriter.java

XML writer class.

------------------------------------------------------------------------

This file is part of NlfExample, a Java application that demonstrates
the use of the uk.org.blankaspect.nlf API.
Copyright 2007-2013 Andy Morgan-Richards.

NlfExample is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package xml;

//----------------------------------------------------------------------


// IMPORTS


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;

import java.nio.charset.Charset;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import util.SystemUtilities;

//----------------------------------------------------------------------


// XML WRITER CLASS


public class XmlWriter
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    public enum Standalone
    {
        NONE,
        NO,
        YES
    }

    public static final     int INDENT_INCREMENT    = 2;

    public static final     String  COMMENT_PREFIX  = "<!--";
    public static final     String  COMMENT_SUFFIX  = "-->";

    private static final    String  DEFAULT_LINE_SEPARATOR  = "\n";

    private static final    String  VERSION_STR     = " version=";
    private static final    String  ENCODING_STR    = " encoding=";
    private static final    String  STANDALONE_STR  = " standalone=";

    private static final    String  DOCTYPE_STR = "<!DOCTYPE ";
    private static final    String  PUBLIC_STR  = "PUBLIC ";
    private static final    String  SYSTEM_STR  = "SYSTEM ";

    private static final    String  NO_STR  = "no";
    private static final    String  YES_STR = "yes";

    private interface AttrName
    {
        String  XMLNS   = "xmlns";
    }

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    public XmlWriter( File   file,
                      String encodingName )
        throws FileNotFoundException, SecurityException, UnsupportedEncodingException
    {
        fileOutStream = new FileOutputStream( file );
        init( new BufferedWriter( new OutputStreamWriter( fileOutStream, encodingName ) ) );
    }

    //------------------------------------------------------------------

    public XmlWriter( File    file,
                      Charset encoding )
        throws FileNotFoundException, SecurityException
    {
        fileOutStream = new FileOutputStream( file );
        init( new BufferedWriter( new OutputStreamWriter( fileOutStream, encoding ) ) );
    }

    //------------------------------------------------------------------

    public XmlWriter( OutputStream outStream,
                      String       encodingName )
        throws UnsupportedEncodingException
    {
        init( new BufferedWriter( new OutputStreamWriter( outStream, encodingName ) ) );
    }

    //------------------------------------------------------------------

    public XmlWriter( OutputStream outStream,
                      Charset      encoding )
    {
        init( new BufferedWriter( new OutputStreamWriter( outStream, encoding ) ) );
    }

    //------------------------------------------------------------------

    public XmlWriter( Writer writer )
    {
        init( writer );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    public static List<Attribute> createAttributeList( Element element )
    {
        List<Attribute> attributes = new ArrayList<Attribute>( );
        NamedNodeMap attrs = element.getAttributes( );
        for ( int i = 0; i < attrs.getLength( ); ++i )
        {
            Node node = attrs.item( i );
            Attribute attribute = new Attribute( node.getNodeName( ), node.getNodeValue( ), true );
            if ( node.getNodeName( ).startsWith( AttrName.XMLNS ) )
                attributes.add( 0, attribute );
            else
                attributes.add( attribute );
        }
        return attributes;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    public FileOutputStream getFileOutStream( )
    {
        return fileOutStream;
    }

    //------------------------------------------------------------------

    public String getLineSeparator( )
    {
        return lineSeparator;
    }

    //------------------------------------------------------------------

    public int getOutLength( )
    {
        return outBuffer.length( );
    }

    //------------------------------------------------------------------

    public void setLineSeparator( String lineSeparator )
    {
        this.lineSeparator = (lineSeparator == null) ? SystemUtilities.getLineSeparator( )
                                                     : lineSeparator;
    }

    //------------------------------------------------------------------

    public void close( )
        throws IOException
    {
        try
        {
            if ( outStream != null )
            {
                outStream.append( outBuffer );
                outStream.close( );
            }
        }
        catch ( IOException e )
        {
            throw e;
        }
        finally
        {
            outStream = null;
            outBuffer.setLength( 0 );
        }
    }

    //------------------------------------------------------------------

    public void writeXmlDeclaration( CharSequence versionStr,
                                     CharSequence encodingName,
                                     Standalone   standalone )
        throws IOException
    {
        write( "<?xml" );
        if ( versionStr != null )
        {
            write( VERSION_STR );
            writeQuoted( versionStr );
        }
        if ( encodingName != null )
        {
            write( ENCODING_STR );
            writeQuoted( encodingName );
        }
        if ( standalone != Standalone.NONE )
        {
            write( STANDALONE_STR );
            writeQuoted( (standalone == Standalone.NO) ? NO_STR : YES_STR );
        }
        write( "?>" );
        writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeDocumentType( CharSequence documentName,
                                   CharSequence systemId,
                                   CharSequence publicId )
        throws IOException
    {
        write( DOCTYPE_STR );
        write( documentName );
        writeLineSeparator( );
        writeSpaces( DOCTYPE_STR.length( ) );
        if ( publicId != null )
        {
            write( PUBLIC_STR );
            writeQuoted( publicId );
            writeLineSeparator( );
            writeSpaces( DOCTYPE_STR.length( ) );
        }
        else
            write( SYSTEM_STR );
        writeQuoted( systemId );
        write( '>' );
        writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeProcessingInstruction( CharSequence target,
                                            CharSequence data )
        throws IOException
    {
        write( "<?" );
        write( target );
        write( ' ' );
        write( data );
        write( "?>" );
        writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeElementStart( CharSequence name,
                                   int          indent,
                                   boolean      elementNewLine )
        throws IOException
    {
        writeElementStart( name, null, indent, elementNewLine, false );
    }

    //------------------------------------------------------------------

    public void writeElementStart( CharSequence    name,
                                   List<Attribute> attributes,
                                   int             indent,
                                   boolean         elementNewLine,
                                   boolean         attrNewLine )
        throws IOException
    {
        writeSpaces( indent );
        write( '<' );
        write( name );
        if ( attributes != null )
            writeAttributes( attributes, attrNewLine ? indent + name.length( ) + 2 : 0 );
        write( '>' );
        if ( elementNewLine )
            writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeEndTag( CharSequence name )
        throws IOException
    {
        write( "</" );
        write( name );
        write( '>' );
    }

    //------------------------------------------------------------------

    public void writeElementEnd( CharSequence name,
                                 int          indent )
        throws IOException
    {
        if ( indent > 0 )
            writeSpaces( indent );
        writeEndTag( name );
        writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeEmptyElement( CharSequence    name,
                                   List<Attribute> attributes,
                                   int             indent,
                                   boolean         attrNewLine )
        throws IOException
    {
        writeSpaces( indent );
        write( '<' );
        write( name );
        if ( attributes != null )
            writeAttributes( attributes, attrNewLine ? indent + name.length( ) + 2 : 0 );
        write( "/>" );
        writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeEscapedTextElement( CharSequence name,
                                         int          indent,
                                         String       text )
        throws IOException
    {
        writeElementStart( name, indent, false );
        writeEscaped( text );
        writeElementEnd( name, 0 );
    }

    //------------------------------------------------------------------

    public void writeComment( CharSequence comment,
                              int          indent )
        throws IOException
    {
        writeSpaces( indent );
        write( COMMENT_PREFIX );
        write( ' ' );
        write( comment );
        write( ' ' );
        write( COMMENT_SUFFIX );
        writeLineSeparator( );
    }

    //------------------------------------------------------------------

    public void writeSpaces( int numSpaces )
        throws IOException
    {
        writeChars( ' ', numSpaces );
    }

    //------------------------------------------------------------------

    public void write( char ch )
        throws IOException
    {
        if ( ch == '\n' )
        {
            outBuffer.append( lineSeparator );
            outStream.append( outBuffer );
            outBuffer.setLength( 0 );
        }
        else
            outBuffer.append( ch );
    }

    //------------------------------------------------------------------

    public void write( CharSequence charSeq )
        throws IOException
    {
        for ( int i = 0; i < charSeq.length( ); ++i )
            write( charSeq.charAt( i ) );
    }

    //------------------------------------------------------------------

    public void writeEscaped( CharSequence charSeq )
        throws IOException
    {
        write( XmlUtilities.escape( charSeq ) );
    }

    //------------------------------------------------------------------

    public void writeQuoted( CharSequence charSeq )
        throws IOException
    {
        write( '"' );
        write( charSeq );
        write( '"' );
    }

    //------------------------------------------------------------------

    public void writeLineSeparator( )
        throws IOException
    {
        write( '\n' );
    }

    //------------------------------------------------------------------

    public void writeChars( char ch,
                            int  count )
        throws IOException
    {
        for ( int i = 0; i < count; ++i )
            write( ch );
    }

    //------------------------------------------------------------------

    public void writeElement( Element element,
                              int     indent )
        throws IOException
    {
        writeElement( element, indent, false );
    }

    //------------------------------------------------------------------

    public void writeElement( Element element,
                              int     indent,
                              boolean attrNewLine )
        throws IOException
    {
        // Create list of attributes
        List<Attribute> attributes = createAttributeList( element );

        // Write element
        if ( element.hasChildNodes( ) )
        {
            // Test for child elements
            boolean hasChildElements = XmlUtilities.hasChildren( element, Node.ELEMENT_NODE );

            // Write element start tag
            writeElementStart( element.getTagName( ), attributes, indent, hasChildElements, attrNewLine );

            // Write child elements
            NodeList nodes = element.getChildNodes( );
            for ( int i = 0; i < nodes.getLength( ); ++i )
            {
                Node node = nodes.item( i );
                if ( node.getNodeType( ) == Node.ELEMENT_NODE )
                    writeElement( (Element)node, indent + INDENT_INCREMENT, attrNewLine );
                else if ( node.getNodeType( ) == Node.TEXT_NODE )
                    write( node.getNodeValue( ) );
            }

            // Write element end tag
            writeElementEnd( element.getTagName( ), hasChildElements ? indent : 0 );
        }
        else
            writeEmptyElement( element.getTagName( ), attributes, indent, attrNewLine );
    }

    //------------------------------------------------------------------

    private void init( Writer writer )
    {
        outStream = writer;
        outBuffer = new StringBuilder( 256 );
        lineSeparator = DEFAULT_LINE_SEPARATOR;
    }

    //------------------------------------------------------------------

    private void writeAttributes( List<Attribute> attributes,
                                  int             indent )
        throws IOException
    {
        for ( int i = 0; i < attributes.size( ); ++i )
        {
            if ( i == 0 )
                write( ' ' );
            else
            {
                if ( indent > 0 )
                {
                    writeLineSeparator( );
                    writeSpaces( indent );
                }
                else
                    write( ' ' );
            }
            Attribute attr = attributes.get( i );
            write( attr.getName( ) );
            write( '=' );
            writeQuoted( attr.getValue( ) );
        }
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    private FileOutputStream    fileOutStream;
    private Writer              outStream;
    private StringBuilder       outBuffer;
    private String              lineSeparator;

}

//----------------------------------------------------------------------
