/*====================================================================*\

Attributes.java

Nested List File attributes chunk class.

------------------------------------------------------------------------

This file is part of the Blank Aspect NLF library (Java package
uk.org.blankaspect.nlf).
Copyright 2007-2013 Andy Morgan-Richards.

The Blank Aspect NLF library is free software: you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package uk.org.blankaspect.nlf;

//----------------------------------------------------------------------


// IMPORTS


import java.io.DataOutput;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.w3c.dom.DOMException;
import org.w3c.dom.Element;

//----------------------------------------------------------------------


// NESTED LIST FILE ATTRIBUTES CHUNK CLASS


/**
 * This class implements an attributes chunk.  An attributes chunk is a special {@linkplain Chunk chunk} in
 * a Nested List File.  A {@linkplain ChunkList chunk list} may contain one attributes chunk, which has the
 * reserved identifier {@code $ATTR} and consists of a list of name-value pairs, analogous to the attributes
 * of an element in an XML document.  The names of the attributes in an attributes list must be unique.  In
 * a Nested List File, the name and value of an attribute are each encoded as two size bytes followed by a
 * UTF-8 sequence.  The byte order of the size bytes is determined by the byte-order flag in the header of
 * the Nested List File {@link Document document} to which the chunk belongs.  The maximum size of a name or
 * value is 65535 bytes.
 * <p>
 * To allow the conversion of a Nested List File to XML, an attribute name must be valid unprefixed name
 * (ie, a name that doesn't contain a ':') under XML 1.1.  Note that XML 1.1 names are less restrictive than
 * those of XML 1.0, so that a name that is valid under XML 1.1 might not be valid under XML 1.0.
 * </p>
 * <p>
 * When an attributes chunk is constructed, its chunk writer is set to a default writer, which is suitable
 * for use by {@link Document#write(java.io.File) Document.write(java.io.File)}.  The default writer may be
 * replaced, using {@link Chunk#setWriter(Chunk.Writer) Chunk.setWriter(Chunk.Writer)}, if it is unsuitable
 * for a particular purpose.  The {@code rewrite} flag, which is accessed with {@link #isRewrite()
 * isRewrite()} and {@link #setRewrite(boolean) setRewrite(boolean)}, is used by the default writer to
 * indicate to the {@linkplain Document#write(java.io.File) document writer} whether the attributes chunk
 * should be rewritten on the second pass (for example, if an attribute value can be set only after
 * subsequent chunks have been written).
 * </p>
 *
 * @since 1.0
 * @see   Chunk
 * @see   ChunkList
 */

public class Attributes
    extends Chunk
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    // Attributes chunk identifier
    public static final     Id  ATTRIBUTES_ID   = new Id( Id.RESERVED_PREFIX + "ATTR" );

    // Size of attribute name
    public static final     int NAME_SIZE_SIZE  = 2;
    public static final     int MIN_NAME_SIZE   = 1;
    public static final     int MAX_NAME_SIZE   = (1 << 16) - 1;
    protected static final  int NAME_SIZE_MASK  = (1 << (NAME_SIZE_SIZE << 3)) - 1;

    // Size of attribute value
    public static final     int VALUE_SIZE_SIZE = 2;
    public static final     int MIN_VALUE_SIZE  = 0;
    public static final     int MAX_VALUE_SIZE  = (1 << 16) - 1;
    protected static final  int VALUE_SIZE_MASK = (1 << (VALUE_SIZE_SIZE << 3)) - 1;

////////////////////////////////////////////////////////////////////////
//  Member classes : non-inner classes
////////////////////////////////////////////////////////////////////////


    // ATTRIBUTE NAME-VALUE PAIR CLASS


    /**
     * This class implements an attribute name-value pair.  An {@linkplain Attributes attributes chunk}
     * consists of a list of name-value pairs.
     *
     * @since 1.0
     */

    public static class Attr
        implements Comparable<Attr>
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs an attribute name-value pair with the specified name.  The implicit attribute value is
         * an empty string.
         *
         * @param  name  the attribute name.
         * @throws IllegalArgumentException
         *           if
         *           <ul>
         *             <li>{@code name} is {@code null}, or</li>
         *             <li>the length of the UTF-8 encoding of {@code name} is greater than 65535 bytes,
         *                 or</li>
         *             <li>{@code name} is not a valid attribute name.</li>
         *           </ul>
         * @since  1.0
         */

        public Attr( String name )
            throws IllegalArgumentException
        {
            this( name, new String( ) );
        }

        //--------------------------------------------------------------

        /**
         * Constructs an attribute name-value pair with the specified name and value.
         *
         * @param  name   the attribute name.
         * @param  value  the attribute value.
         * @throws IllegalArgumentException
         *           if
         *           <ul>
         *             <li>{@code name} is {@code null}, or</li>
         *             <li>the length of the UTF-8 encoding of {@code name} is greater than 65535 bytes,
         *                 or</li>
         *             <li>{@code name} is not a valid attribute name, or</li>
         *             <li>{@code value} is {@code null}, or</li>
         *             <li>the length of the UTF-8 encoding of {@code value} is greater than 65535
         *                 bytes.</li>
         *           </ul>
         * @since  1.0
         */

        public Attr( String name,
                     String value )
            throws IllegalArgumentException
        {
            // Validate name
            if ( (name == null) || !isValidName( name ) )
                throw new IllegalArgumentException( );

            // Validate value
            if ( (value == null) ||
                 !NlfUtilities.isUtf8LengthWithinBounds( value, MIN_VALUE_SIZE, MAX_VALUE_SIZE ) )
                throw new IllegalArgumentException( );

            // Set name and value
            this.name = name;
            this.value = value;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Comparable interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Compares this attribute with the specified {@code Attr} object.  The comparison is performed by
         * comparing the names of the two objects using {@code
         * java.lang.String.compareTo(java.lang.String)}, which compares strings lexicographically by the
         * Unicode value of each character in the strings.
         * <p>
         * The result is
         * </p>
         * <ul>
         *   <li>a negative integer if the name of this attribute lexicographically precedes the name of the
         *       argument;</li>
         *   <li>a positive integer if the name of this attribute lexicographically succeeds the name of the
         *       argument;</li>
         *   <li>zero if the names of the two attributes are equal.</li>
         * </ul>
         * <p>
         * Note that a result of zero <em>does not</em> imply that the {@link #equals(Object)
         * equals(Object)} method would return {@code true}, although the converse is true.
         * </p>
         *
         * @param  attr  the {@code Attr} object with which the comparison is to be made.
         * @return <ul>
         *           <li>{@code 0} (zero) if the name of this attribute is equal to the name of {@code
         *               attr};</li>
         *           <li>a value less than {@code 0} if the name of this attribute is lexicographically less
         *               than the name of {@code attr};</li>
         *           <li>a value greater than {@code 0} if the name of this attribute is lexicographically
         *               greater than the name of {@code attr}.</li>
         *         </ul>
         * @since  1.0
         * @see    java.lang.String#compareTo(java.lang.String)
         */

        public int compareTo( Attr attr )
        {
            return name.compareTo( attr.name );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : overriding methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Compares this attribute with the specified object.  The result is {@code true} if and only if
         * {@code obj} is an {@code Attr} object that has the same name and value as this object.
         *
         * @param  obj  the object with which the comparison is to be made.
         * @return {@code true} if {@code obj} is an {@code Attr} object with the same name and value as
         *         this object; {@code false} otherwise.
         * @since  1.0
         */

        @Override
        public boolean equals( Object obj )
        {
            if ( obj instanceof Attr )
            {
                Attr attr = (Attr)obj;
                return ( name.equals( attr.name ) && value.equals( attr.value ) );
            }
            return false;
        }

        //--------------------------------------------------------------

        /**
         * Returns the hash code for this object.
         *
         * @return the hash code for this object.
         * @since  1.1
         * @see    #equals(Object)
         */

        @Override
        public int hashCode( )
        {
            return ( name.hashCode( ) * 31 + value.hashCode( ) );
        }

        //--------------------------------------------------------------

        /**
         * Returns the attribute as a string.  The string is in the form <i>name</i>="<i>value</i>".
         *
         * @return the attribute as a string, in the form <i>name</i>="<i>value</i>".
         * @since  1.0
         */

        @Override
        public String toString( )
        {
            return ( name + "=\"" + value + "\"" );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Returns the name of the attribute.
         *
         * @return the name of the attribute.
         * @since  1.0
         */

        public String getName( )
        {
            return name;
        }

        //--------------------------------------------------------------

        /**
         * Returns the value of the attribute.
         *
         * @return the value of the attribute.
         * @since  1.0
         */

        public String getValue( )
        {
            return value;
        }

        //--------------------------------------------------------------

        /**
         * Returns the size of this {@code Attr} object in its encoded form (ie, in the form in which it is
         * encoded in an attributes chunk in a Nested List File).  The name and value of the attribute are
         * each encoded as two size bytes followed by a UTF-8 sequence.
         *
         * @return the size of the {@code Attr} object in its encoded form.
         * @since  1.0
         * @see    #getBytes(boolean)
         */

        public int getSize( )
        {
            return ( NAME_SIZE_SIZE + NlfUtilities.getUtf8Length( name ) + VALUE_SIZE_SIZE +
                                                                    NlfUtilities.getUtf8Length( value ) );
        }

        //--------------------------------------------------------------

        /**
         * Returns this {@code Attr} object in its encoded form (ie, in the form in which it is encoded in
         * an attributes chunk in a Nested List File).  The name and value of the attribute are each encoded
         * as two size bytes followed by a UTF-8 sequence.
         *
         * @param  littleEndian  {@code true} if the byte order of the size of the name and value is
         *                       little-endian; {@code false} if the byte order of the size of the name and
         *                       value is big-endian.
         * @return the {@code Attr} object in its encoded form.
         * @since  1.0
         * @see    #getSize()
         */

        public byte[] getBytes( boolean littleEndian )
        {
            byte[] nameBytes = NlfUtilities.stringToUtf8( name );
            byte[] nameSizeBuffer = new byte[NAME_SIZE_SIZE];
            Util.intToBytes( nameBytes.length, nameSizeBuffer, 0, nameSizeBuffer.length, littleEndian );

            byte[] valueBytes = NlfUtilities.stringToUtf8( value );
            byte[] valueSizeBuffer = new byte[VALUE_SIZE_SIZE];
            Util.intToBytes( valueBytes.length, valueSizeBuffer, 0, valueSizeBuffer.length, littleEndian );

            int size = NAME_SIZE_SIZE + nameBytes.length + VALUE_SIZE_SIZE + valueBytes.length;
            byte[] buffer = new byte[size];
            int offset = 0;
            System.arraycopy( nameSizeBuffer, 0, buffer, offset, nameSizeBuffer.length );
            offset += nameSizeBuffer.length;
            System.arraycopy( nameBytes, 0, buffer, offset, nameBytes.length );
            offset += nameBytes.length;
            System.arraycopy( valueSizeBuffer, 0, buffer, offset, valueSizeBuffer.length );
            offset += valueSizeBuffer.length;
            System.arraycopy( valueBytes, 0, buffer, offset, valueBytes.length );

            return buffer;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private String  name;
        private String  value;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Member classes : inner classes
////////////////////////////////////////////////////////////////////////


    // ATTRIBUTES WRITER CLASS


    /**
     * This class implements a {@linkplain Chunk.Writer chunk writer}, with which the list of attributes in
     * an attributes chunk is written to a Nested List File.  An instance of this class is set as the
     * default writer when an {@link Attributes} object is created.
     *
     * @since 1.0
     */

    private class Writer
        implements Chunk.Writer
    {

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs an attributes writer, which implements the {@link Chunk.Writer Chunk.Writer}
         * interface.
         *
         * @since 1.0
         */

        private Writer( )
        {
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Chunk.Writer interface
    ////////////////////////////////////////////////////////////////////

        /**
         * This method, which is called before a chunk is written, performs no action.  It returns the value
         * of the {@link Attributes#rewrite Attributes.rewrite} field to indicate whether the chunk should
         * be rewritten on the second pass of the document writer.
         *
         * @param  pass  the index of the pass (0 = first pass, 1 = second pass) of the document writer that
         *               calls this method.
         * @return {@code true} if the chunk should be rewritten on the second pass of the document writer;
         *         {@code false} otherwise.
         * @since  1.0
         * @see    Chunk.Writer#reset(int)
         * @see    Document#write(java.io.File)
         * @see    Attributes#rewrite
         */

        public boolean reset( int pass )
        {
            return ( (pass == 0) ? rewrite : false );
        }

        //--------------------------------------------------------------

        /**
         * Returns the size of the attributes chunk.
         *
         * @return the size of the attributes chunk.
         * @since  1.0
         * @see    Chunk.Writer#getLength( int )
         * @see    Document#write(java.io.File)
         */

        public long getLength( )
        {
            long size = 0;
            for ( Attr attr : attributes )
                size += attr.getSize( );
            return size;
        }

        //--------------------------------------------------------------

        /**
         * Writes the attributes chunk to the specified output.
         *
         * @param  dataOutput  the {@code java.io.DataOutput} object to which the attributes chunk is to be
         *                     written.
         * @throws IOException
         *           if an I/O error occurs.
         * @since  1.0
         * @see    Chunk.Writer#write(java.io.DataOutput)
         * @see    Document#write(java.io.File)
         */

        public void write( DataOutput dataOutput )
            throws IOException
        {
            for ( Attr attr : attributes )
                dataOutput.write( attr.getBytes( document.isLittleEndian( ) ) );
        }

        //--------------------------------------------------------------

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    /**
     * Constructs an attributes chunk.  An attributes chunk can be constructed only by the {@linkplain
     * ChunkList list} to which it belongs.  A list can contain no more than one attributes chunk.
     * <p>
     * The chunk writer for this chunk is set to a default chunk writer when an {@code Attributes} object is
     * constructed.  The default writer may be replaced by the base-class method {@link Chunk#setWriter(
     * Chunk.Writer ) Chunk.setWriter(Chunk.Writer)}.
     * </p>
     *
     * @param document  the document to which this chunk is to belong.
     * @since 1.0
     */

    protected Attributes( Document document )
    {
        super( document, ATTRIBUTES_ID );
        size = 0;
        writer = new Writer( );
        attributes = new ArrayList<Attr>( );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    /**
     * Tests whether the specified string is a valid attribute name.  An attribute name is valid if it is a
     * valid unprefixed name (ie, a name that doesn't contain a ':') under XML 1.1.
     *
     * @param  str  the string that is to be tested.
     * @return {@code true} if the specified string is a valid attribute name; {@code false} otherwise.
     * @since  1.0
     */

    public static boolean isValidName( String str )
    {
        if ( !NlfUtilities.isUtf8LengthWithinBounds( str, MIN_NAME_SIZE, MAX_NAME_SIZE ) )
            return false;

        int index = 0;
        while ( index < str.length( ) )
        {
            int codePoint = str.codePointAt( index );
            if ( index == 0 )
            {
                if ( !NlfUtilities.isNameStartChar( codePoint ) )
                    return false;
            }
            else
            {
                if ( !NlfUtilities.isNameChar( codePoint ) )
                    return false;
            }
            index += Character.charCount( codePoint );
        }
        return true;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods : overriding methods
////////////////////////////////////////////////////////////////////////

    /**
     * Compares this attributes chunk with a specified object.  The result is {@code true} if and only if
     * {@code obj} is an {@code Attributes} object that contains the same attributes as this object, though
     * not necessarily in the same order.
     *
     * @param  obj  the object with which the comparison is to be made.
     * @return {@code true} if {@code obj} is an {@code Attributes} object that contains the same attributes
     *         as this object, though not necessarily in the same order; {@code false} otherwise.
     * @since  1.0
     */

    @Override
    public boolean equals( Object obj )
    {
        if ( obj instanceof Attributes )
        {
            Attributes attrs = (Attributes)obj;
            if ( attributes.size( ) == attrs.attributes.size( ) )
            {
                for ( Attr attr : attributes )
                {
                    if ( !attrs.attributes.contains( attr ) )
                        return false;
                }
                return true;
            }
        }
        return false;
    }

    //------------------------------------------------------------------

    /**
     * Returns the hash code for this object.
     *
     * @return the hash code for this object.
     * @since  1.1
     * @see    #equals(Object)
     */

    @Override
    public int hashCode( )
    {
        Attr[] attrArray = attributes.toArray( new Attr[attributes.size( )] );
        Arrays.sort( attrArray );
        return Arrays.hashCode( attrArray );
    }

    //------------------------------------------------------------------

    /**
     * Returns a string representation of the attributes chunk.  Each attribute in the attributes list is in
     * the form <i>name</i>="<i>value</i>", and attributes are separated with a comma and a space.
     *
     * @return the attributes chunk as a string, with each attribute in the form
     *         <i>name</i>="<i>value</i>", and attributes separated with a comma and a space.
     * @since  1.0
     * @see    Attr#toString()
     */

    @Override
    public String toString( )
    {
        StringBuilder buffer = new StringBuilder( );
        for ( int i = 0; i < attributes.size( ); ++i )
        {
            if ( i > 0 )
                buffer.append( ", " );
            buffer.append( attributes.get( i ) );
        }
        return buffer.toString( );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    /**
     * Returns {@code true} if the attributes will be rewritten on the second pass of the document writer.
     *
     * @return {@code true} if the attributes will be rewritten on the second pass of the document writer;
     *         {@code false} otherwise.
     * @since  1.0
     * @see    #setRewrite(boolean)
     */

    public boolean isRewrite( )
    {
        return rewrite;
    }

    //------------------------------------------------------------------

    /**
     * Returns {@code true} if this chunk contains at least one attribute.
     *
     * @return {@code true} if this chunk contains at least one attribute; {@code false}
     *         otherwise.
     * @since  1.0
     * @see    #getNumAttributes()
     */

    public boolean hasAttributes( )
    {
        return !attributes.isEmpty( );
    }

    //------------------------------------------------------------------

    /**
     * Returns the number of attributes in this chunk's attributes list.
     *
     * @return the number of attributes in this chunk's attributes list.
     * @since  1.0
     * @see    #hasAttributes()
     */

    public int getNumAttributes( )
    {
        return attributes.size( );
    }

    //------------------------------------------------------------------

    /**
     * Returns the attribute at the specified index in the attributes list.
     *
     * @return the the attribute at the specified index in the attributes list.
     * @throws IndexOutOfBoundsException
     *           if {@code (index < 0)} or {@code (index >= }{@link #getNumAttributes()
     *           getNumAttributes()}{@code )}.
     * @since  1.0
     * @see    #getNumAttributes()
     */

    public Attr getAttribute( int index )
        throws IndexOutOfBoundsException
    {
        return attributes.get( index );
    }

    //------------------------------------------------------------------

    /**
     * Sets the flag that indicates whether the attributes will be rewritten on the second pass of the
     * document writer.  The flag, which is used by an {@code Attributes} object's default writer, can be
     * tested by a replacement writer with {@link #isRewrite()}.
     *
     * @param rewrite  {@code true} if the attributes should be rewritten on the second pass of the document
     *                 writer; {@code false} otherwise.
     * @since 1.0
     * @see   #isRewrite()
     */

    public void setRewrite( boolean rewrite )
    {
        this.rewrite = rewrite;
    }

    //------------------------------------------------------------------

    /**
     * Returns the index of the attribute with the specified name in this chunk's attributes list, or {@code
     * -1} if the list does not contain an attribute with the specified name.  As the attributes in the list
     * must have unique names, any occurrence in the list of an attribute with a given name will be the only
     * occurrence.
     *
     * @param  name  the name of the attribute whose index is to be determined.
     * @return the index of the attribute with the specified name the attributes list, or {@code -1} if the
     *         list does not contain an attribute with the specified name.
     * @throws IllegalArgumentException
     *           if {@code name} is {@code null}.
     * @since  1.0
     */

    public int indexOf( String name )
        throws IllegalArgumentException
    {
        if ( name == null )
            throw new IllegalArgumentException( );

        for ( int i = 0; i < attributes.size( ); ++i )
        {
            if ( attributes.get( i ).name.equals( name ) )
                return i;
        }
        return -1;
    }

    //------------------------------------------------------------------

    /**
     * Sets the specified attribute in this chunk's attributes list.  If there is already an attribute with
     * the same name in the list, it is replaced with the specified attribute; otherwise, the new attribute
     * is added to the end of the list.
     *
     * @param  attribute  the attribute that is to be set in or added to the list of attributes.
     * @throws IllegalArgumentException
     *           if {@code attribute} is {@code null}.
     * @since  1.0
     * @see    #removeAttribute(int)
     */

    public void setAttribute( Attr attribute )
        throws IllegalArgumentException
    {
        // Validate argument
        if ( attribute == null )
            throw new IllegalArgumentException( );

        // Get index to list
        int index = indexOf( attribute.name );

        // Add attribute to list
        if ( index < 0 )
            attributes.add( attribute );

        // Replace attribute in list
        else
        {
            size -= attributes.get( index ).getSize( );
            attributes.set( index, attribute );
        }

        // Increment chunk size
        size += attribute.getSize( );
    }

    //------------------------------------------------------------------

    /**
     * Removes the attribute at the specified index from this chunk's attributes list, and returns the
     * attribute that was removed.
     *
     * @param  index  the index of the attribute that is to be removed from the list of attributes.
     * @return the attribute that was removed from the list.
     * @throws IndexOutOfBoundsException
     *           if {@code (index < 0)} or {@code (index >= }{@link #getNumAttributes()
     *           getNumAttributes()}{@code )}.
     * @since  1.0
     * @see    #setAttribute(Attr)
     * @see    #getNumAttributes()
     * @see    #indexOf(java.lang.String)
     */

    public Attr removeAttribute( int index )
        throws IndexOutOfBoundsException
    {
        Attr attr = attributes.remove( index );
        size -= attr.getSize( );
        return attr;
    }

    //------------------------------------------------------------------

    /**
     * Sets the attributes in this chunk's attributes list as attributes of the specified XML element.
     *
     * @param  element  the XML element on which the attributes are to be set.
     * @throws DOMException
     *           if an attribute name is not valid for the XML version of the document to which {@code
     *           element} belongs.
     * @since  1.0
     */

    protected void toXml( Element element )
        throws DOMException
    {
        for ( Attr attribute : attributes )
            element.setAttribute( attribute.name, attribute.value );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    private List<Attr>  attributes;
    private boolean     rewrite;

}

//----------------------------------------------------------------------
