/*====================================================================*\

Chunk.java

Nested List File chunk class.

------------------------------------------------------------------------

This file is part of the Blank Aspect NLF library (Java package
uk.org.blankaspect.nlf).
Copyright 2007-2013 Andy Morgan-Richards.

The Blank Aspect NLF library is free software: you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package uk.org.blankaspect.nlf;

//----------------------------------------------------------------------


// IMPORTS


import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import org.w3c.dom.DOMException;
import org.w3c.dom.Element;

//----------------------------------------------------------------------


// NESTED LIST FILE CHUNK CLASS


/**
 * This class implements a chunk of a Nested List File.  The chunk, which is the primary structural unit of
 * a Nested List File, consists of a header and binary data.  The header consists of an {@linkplain Id
 * identifier} and a size field.  The identifier field is of variable size, up to 256 bytes; the size of the
 * size field is 8 bytes.  The chunk size does not include the size of the header.
 * <p>
 * A chunk is created with the {@link Document#createChunk(Id) Document.createChunk(Id)} method, and is
 * added to a list with the {@link ChunkList#appendChunk(Chunk) ChunkList.appendChunk(Chunk)} method.  A
 * chunk can be added only to a list that belongs to the same document as the chunk.
 * </p>
 * <p>
 * There are two special chunks: the {@linkplain ChunkList list} and the {@linkplain Attributes attributes
 * chunk}, which have the reserved identifiers {@code "$LIST"} and {@code "$ATTR"} respectively.
 * </p>
 *
 * @since 1.0
 * @see   ChunkList
 */

public class Chunk
    implements Comparable<Chunk>
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    // Size of chunk
    public static final int     SIZE_SIZE   = 8;
    public static final long    MIN_SIZE    = 0;
    public static final long    MAX_SIZE    = (1L << 62) - 1;

////////////////////////////////////////////////////////////////////////
//  Member interfaces
////////////////////////////////////////////////////////////////////////


    // CHUNK-DATA READER INTERFACE


    /**
     * This interface specifies the methods that must be implemented by a class that reads chunk data from a
     * data input.  The data input is likely to be an instance of {@code java.io.RandomAccessFile}.
     *
     * @since 1.0
     * @see   #getReader()
     * @see   #setReader(Chunk.Reader)
     */

    public interface Reader
    {

    ////////////////////////////////////////////////////////////////////
    //  Methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk reader before any reading is performed.
         *
         * @throws IOException
         *           if an I/O error occurs.
         * @since  1.0
         */

        void reset( )
            throws IOException;

        //--------------------------------------------------------------

        /**
         * Returns the data input from which the chunk data are to be read.  The input is likely to be an
         * instance of {@code java.io.RandomAccessFile}.
         *
         * @return the data input from which the chunk data are to be read.
         * @since  1.0
         */

        DataInput getDataInput( );

        //--------------------------------------------------------------

    }

    //==================================================================


    // CHUNK-DATA WRITER INTERFACE


    /**
     * This interface specifies the methods that must be implemented by a class that writes chunk data to a
     * data output.  The data output is likely to be an instance of {@code java.io.RandomAccessFile}.
     * <p>
     * The {@link Document#write(java.io.File) Document.write(java.io.File)} method, which is provided as a
     * standard way of writing a Nested List File, writes a document in two passes.  It uses the value that
     * is returned by the {@link #reset(int) reset(int)} method on the first pass to determine whether a
     * chunk should be rewritten on the second pass.
     * </p>
     *
     * @since 1.0
     * @see   #getWriter()
     * @see   #setWriter(Chunk.Writer)
     */

    public interface Writer
    {

    ////////////////////////////////////////////////////////////////////
    //  Methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk writer before any writing is performed.  The writer is reset at the start of
         * each pass of the document writer.
         *
         * @param  pass  the index of the pass (0 = first pass, 1 = second pass) of the document writer that
         *               calls this method.
         * @return {@code true} on the first pass if the chunk should be rewritten on the second pass;
         *         {@code false} otherwise.
         * @throws IOException
         *           if an I/O error occurs.
         * @since  1.0
         */

        boolean reset( int pass )
            throws IOException;

        //--------------------------------------------------------------

        /**
         * Returns the length (size) of the chunk in bytes.
         *
         * @return the length (size) of the chunk in bytes.
         * @since  1.0
         */

        long getLength( );

        //--------------------------------------------------------------

        /**
         * Writes the chunk data to the specified data output.  The data output is likely to be an instance
         * of {@code java.io.RandomAccessFile}.
         *
         * @param  dataOutput  the data output to which the chunk data are to be written.
         * @throws IOException
         *           if an I/O error occurs.
         * @since  1.0
         */

        void write( DataOutput dataOutput )
            throws IOException;

        //--------------------------------------------------------------

    }

    //==================================================================


    // CHUNK-DATA ENCODER INTERFACE


    /**
     * This interface specifies the methods that must be implemented by a class that encodes chunk data as
     * character data for use as a text node of an XML file.  The encoder is used by the chunk's {@link
     * #toXml(org.w3c.dom.Document) toXml(org.w3c.dom.Document)} method.
     *
     * @since 1.0
     * @see   #getEncoder()
     * @see   #setEncoder(Chunk.Encoder)
     */

    public interface Encoder
    {

    ////////////////////////////////////////////////////////////////////
    //  Methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk encoder before any encoding is performed.
         *
         * @since 1.0
         */

        void reset( );

        //--------------------------------------------------------------

        /**
         * Returns the length of a block of input data that the {@link #encode encode} method expects to
         * receive.  A value of zero indicates that the {@code encode} method will accept a block of any
         * length.
         *
         * @param  size  the size of the chunk data.
         * @return the length of a block of input data that the {@link #encode encode} method expects to
         *         receive; {@code 0}, if the {@code encode} method will accept a block of any length.
         * @since  1.0
         */

        int getInputLength( long size );

        //--------------------------------------------------------------

        /**
         * Encodes the specified byte data as a string.  When the encoder is used by the chunk's {@link
         * #toXml(org.w3c.dom.Document) toXml(org.w3c.dom.Document)} method, this method will be called on
         * successive blocks of the chunk data.
         *
         * @param  data       the data that are to be encoded.
         * @param  offset     the offset to {@code data} at which the input data begin.
         * @param  length     the length of the input data.
         * @param  endOfInput {@code true} if the input data are the last that are to be encoded; {@code
         *                    false} otherwise.
         * @return the string that results from encoding the input data.
         * @throws IllegalArgumentException
         *           if
         *           <ul>
         *             <li>any of the arguments are invalid, or</li>
         *             <li>the input data are malformed or otherwise illegal.</li>
         *           </ul>
         * @since  1.0
         */

        String encode( byte[]  data,
                       int     offset,
                       int     length,
                       boolean endOfInput )
            throws IllegalArgumentException;

        //--------------------------------------------------------------

    }

    //==================================================================


    // CHUNK PROCESSOR INTERFACE


    /**
     * This interface specifies the method that must be implemented by a class that processes the chunks
     * that are visited in the traversal of a document tree or subtree.
     *
     * @since 1.0
     * @see   ChunkList#processChunks(NlfConstants.TraversalOrder, Chunk.Processor)
     * @see   Document#processChunks(NlfConstants.TraversalOrder, Chunk.Processor)
     */

    public interface Processor
    {

    ////////////////////////////////////////////////////////////////////
    //  Methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Processes the specified chunk.  During the traversal of a document tree or subtree, this method
         * is called on every chunk that is visited.
         * <p>
         * A processor that wants to terminate the traversal of the tree should throw a {@link
         * TerminatedException TerminatedException}, which can wrap another {@code java.lang.Throwable}.
         * </p>
         *
         * @param  chunk  the chunk that is to be processed.
         * @throws TerminatedException
         *           if the traversal of the tree or subtree is terminated.
         * @since  1.0
         * @see    TerminatedException
         */

        void process( Chunk chunk )
            throws TerminatedException;

        //--------------------------------------------------------------

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Member classes : non-inner classes
////////////////////////////////////////////////////////////////////////


    // BASE64 ENCODER CLASS


    /**
     * This singleton class implements a Base64 encoder for chunk data.  It is set as the default encoder
     * on all general chunks (ie, chunks that do not have a reserved identifier).  The length of a line of
     * output from the encoder and the string that terminates each line can both be set on the single
     * instance of the class.  The default line length of zero means that no line separators will be
     * appended to the encoded output.
     * <p>
     * A chunk's encoder is used by the {@link #toXml(org.w3c.dom.Document) toXml(org.w3c.dom.Document)}
     * method to encode the chunk data as the character data of an XML element.
     * </p>
     *
     * @since 1.0
     * @see   Encoder
     */

    public static class Base64Encoder
        implements Encoder
    {

    ////////////////////////////////////////////////////////////////////
    //  Constants
    ////////////////////////////////////////////////////////////////////

        private static final    int MIN_LINE_LENGTH = 0;
        private static final    int MAX_LINE_LENGTH = Integer.MAX_VALUE;

        private static final    String  LINE_SEPARATOR_PROPERTY_KEY = "line.separator";

        private static final    String  DEFAULT_LINE_SEPARATOR  = "\n";

        private static final    String  BASE64_CHARS    =
                                        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

    ////////////////////////////////////////////////////////////////////
    //  Constructors
    ////////////////////////////////////////////////////////////////////

        /**
         * Constructs a {@code Base64Encoder}.  This is a singleton class: only one instance of the class
         * is constructed, by {@link #getInstance( ) getInstance( )}.
         *
         * @since 1.0
         */

        private Base64Encoder( )
        {
            setLineSeparator( null );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Class methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Returns the single instance of {@code Base64Encoder}.
         *
         * @return the single instance of {@code Base64Encoder}.
         * @since  1.0
         */

        public static Base64Encoder getInstance( )
        {
            if ( instance == null )
                instance = new Base64Encoder( );
            return instance;
        }

        //--------------------------------------------------------------

        /**
         * Encodes three bytes as four Base64 characters.
         *
         * @param  data  the three bytes that are to be encoded.
         * @return an array of four characters that is the result of encoding the three input bytes as
         *         Base64.
         * @since  1.0
         */

        private static char[] getChars( int data )
        {
            char[] chars = new char[4];
            for ( int j = 3; j >= 0; --j )
            {
                chars[j] = BASE64_CHARS.charAt( data & 0x3F );
                data >>>= 6;
            }
            return chars;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods : Encoder interface
    ////////////////////////////////////////////////////////////////////

        /**
         * Resets the chunk encoder before any encoding is performed: does nothing.
         *
         * @since 1.0
         */

        public void reset( )
        {
            // do nothing
        }

        //--------------------------------------------------------------

        /**
         * Returns the length of a block of input data that the {@link #encode(byte[], int, int, boolean)
         * encode(byte[], int, int, boolean)} method expects to receive.
         *
         * @param  size  the size of the chunk data.
         * @return the length of a block of input data that the {@code encode(byte[], int, int, boolean)}
         *         method expects to receive.
         * @since  1.0
         */

        public int getInputLength( long size )
        {
            return ( lineLength / 4 * 3 );
        }

        //--------------------------------------------------------------

        /**
         * Encodes the specified byte data as a Base64 string.  The line length and line separator that are
         * used in the encoding can be set with {@link #setLineLength(int) setLineLength(int)} and {@link
         * #setLineSeparator(java.lang.String) setLineSeparator(java.lang.String)} respectively.
         *
         * @param  data       the data that are to be encoded.
         * @param  offset     the offset to {@code data} at which the input data begin.
         * @param  length     the length of the input data.
         * @param  endOfInput {@code true} if the input data are the last that are to be encoded; {@code
         *                    false} otherwise.
         * @return the string that results from encoding the input data.
         * @throws IllegalArgumentException
         *           if
         *           <ul>
         *             <li>{@code data} is {@code null}, or</li>
         *             <li>{@code (offset < 0)} or {@code (offset > data.length)}, or</li>
         *             <li>{@code (length < 0)} or {@code (length > data.length - offset)}.</li>
         *           </ul>
         * @since  1.0
         */

        public String encode( byte[]  data,
                              int     offset,
                              int     length,
                              boolean endOfInput )
            throws IllegalArgumentException
        {
            // Validate arguments
            if ( (data == null) || (offset < 0) || (offset > data.length) ||
                 (length < 0) || (length > data.length - offset) )
                throw new IllegalArgumentException( );

            // Encode input data as Base64
            StringBuilder outBuffer = new StringBuilder( 4 * length / 3 );
            int inBuffer = 0;
            int inBufferLength = 0;
            int numChars = 0;
            int endOffset = offset + length;
            while ( offset < endOffset )
            {
                inBuffer <<= 8;
                inBuffer |= data[offset++] & 0xFF;
                if ( ++inBufferLength >= 3 )
                {
                    numChars = appendChars( outBuffer, getChars( inBuffer ), numChars );
                    inBuffer = 0;
                    inBufferLength = 0;
                }
            }
            if ( inBufferLength > 0 )
            {
                inBuffer <<= (3 - inBufferLength) << 3;
                char[] chars = getChars( inBuffer );
                for ( int i = inBufferLength; i < chars.length; ++i )
                    chars[i] = '=';
                appendChars( outBuffer, chars, numChars );
            }
            if ( lineLength > 0 )
                outBuffer.append( lineSeparator );
            return outBuffer.toString( );
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Instance methods
    ////////////////////////////////////////////////////////////////////

        /**
         * Returns the length of a line of encoded output.
         *
         * @return the length of a line of encoded output.
         * @since  1.0
         * @see    #setLineLength(int)
         */

        public int getLineLength( )
        {
            return lineLength;
        }

        //--------------------------------------------------------------

        /**
         * Returns the string that is appended to each line of encoded output.
         *
         * @return the string that is appended to each line of encoded output.
         * @since  1.0
         * @see    #setLineSeparator(String)
         */

        public String getLineSeparator( )
        {
            return lineSeparator;
        }

        //--------------------------------------------------------------

        /**
         * Sets the length of a Base64-encoded line.  If the length is set to zero, no line separators will
         * be appended to the encoded output.  The length must be a multiple of {@code 4}.
         *
         * @param  lineLength  the length of a Base64-encoded line, or {@code 0} if no line separators are
         *                     to be appended to the encoded output.  The length must be a multiple of
         *                     {@code 4}.
         * @throws IllegalArgumentException
         *           if {@code lineLength} is negative or is not a multiple of {@code 4}.
         * @since  1.0
         * @see    #getLineLength()
         */

        public void setLineLength( int lineLength )
            throws IllegalArgumentException
        {
            if ( (lineLength < MIN_LINE_LENGTH) || (lineLength > MAX_LINE_LENGTH) || (lineLength % 4 != 0) )
                throw new IllegalArgumentException( );
            this.lineLength = lineLength;
        }

        //--------------------------------------------------------------

        /**
         * Sets the string that will be appended to each line of encoded output.
         *
         * @param lineSeparator  the string that is to be appended to each line of encoded output.  If
         *                       {@code lineSeparator} is {@code null}, the line separator is set to the
         *                       value of the system property {@code line.separator}.
         * @since 1.0
         * @see   #getLineSeparator()
         */

        public void setLineSeparator( String lineSeparator )
        {
            this.lineSeparator = (lineSeparator == null) ? System.getProperty( LINE_SEPARATOR_PROPERTY_KEY,
                                                                               DEFAULT_LINE_SEPARATOR )
                                                         : lineSeparator;
        }

        //--------------------------------------------------------------

        /**
         * Appends characters to the specified buffer.
         *
         * @param  buffer    the buffer to which the characters are to be appended.
         * @param  chars     the characters that are to be appended.
         * @param  numChars  the number of encoded characters in the buffer.
         * @return the updated number of encoded characters in the buffer.
         * @since  1.0
         */

        private int appendChars( StringBuilder buffer,
                                 char[]        chars,
                                 int           numChars )
        {
            if ( lineLength == 0 )
            {
                buffer.append( chars );
                numChars += chars.length;
            }
            else
            {
                for ( char ch : chars )
                {
                    if ( (numChars > 0) && (numChars % lineLength == 0) )
                        buffer.append( lineSeparator );
                    buffer.append( ch );
                    ++numChars;
                }
            }
            return numChars;
        }

        //--------------------------------------------------------------

    ////////////////////////////////////////////////////////////////////
    //  Class variables
    ////////////////////////////////////////////////////////////////////

        private static  Base64Encoder   instance;

    ////////////////////////////////////////////////////////////////////
    //  Instance variables
    ////////////////////////////////////////////////////////////////////

        private int     lineLength;
        private String  lineSeparator;

    }

    //==================================================================

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    /**
     * Constructs a chunk with the specified owner document and identifier.  A Base64 encoder is set as the
     * default encoder.
     *
     * @param document  the {@linkplain Document document} to which the chunk will belong.
     * @param id        the chunk identifier.
     * @since 1.0
     */

    protected Chunk( Document document,
                     Id       id )
    {
        this.document = document;
        this.id = id;
        size = -1;
        if ( !id.isReserved( ) )
            encoder = Base64Encoder.getInstance( );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods : Comparable interface
////////////////////////////////////////////////////////////////////////

    /**
     * Compares this chunk with the specified {@code Chunk} object.  The comparison is performed by
     * comparing the identifiers of the two objects using {@code
     * java.lang.String.compareTo(java.lang.String)}, which compares strings lexicographically by the
     * Unicode value of each character in the strings.
     * <p>
     * The result is
     * </p>
     * <ul>
     *   <li>a negative integer if the identifier of this chunk lexicographically precedes the identifier of
     *       the argument;</li>
     *   <li>a positive integer if the identifier of this chunk lexicographically succeeds the identifier of
     *       the argument;</li>
     *   <li>zero if the identifiers of the two chunks are equal.</li>
     * </ul>
     * <p>
     * Because of the restrictions on identifiers, special chunks (ie, chunks with reserved identifiers)
     * will always precede general chunks (ie, chunks with non-reserved identifiers).
     * </p>
     *
     * @param  chunk  the {@code Chunk} object with which the comparison is to be made.
     * @return <ul>
     *           <li>{@code 0} (zero) if the identifier of this chunk is equal to the identifier of {@code
     *               chunk};</li>
     *           <li>a value less than {@code 0} if the identifier of this chunk is lexicographically less
     *               than the identifier of {@code chunk};</li>
     *           <li>a value greater than {@code 0} if the identifier of this chunk is lexicographically
     *               greater than the identifier of {@code chunk}.</li>
     *         </ul>
     * @since  1.0
     * @see    Id#compareTo(Id)
     */

    public int compareTo( Chunk chunk )
    {
        return id.compareTo( chunk.id );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    /**
     * Returns the Nested List File document to which this chunk belongs.
     *
     * @return the document to which this chunk belongs.
     * @since  1.0
     */

    public Document getDocument( )
    {
        return document;
    }

    //------------------------------------------------------------------

    /**
     * Returns the chunk list that contains this chunk, or {@code null} if this chunk has no parent.
     *
     * @return the chunk {@linkplain ChunkList list} that contains this chunk, or {@code null} if this chunk
     *         has no parent.
     * @since  1.0
     * @see    #setParent(ChunkList)
     */

    public ChunkList getParent( )
    {
        return parent;
    }

    //------------------------------------------------------------------

    /**
     * Returns the identifier of this chunk.
     *
     * @return the identifier of this chunk.
     * @since  1.0
     */

    public Id getId( )
    {
        return id;
    }

    //------------------------------------------------------------------

    /**
     * Returns the size of this chunk in bytes.  A value of {@code -1} indicates that the size is invalid
     * and that the size of the chunk should be obtained by another means (eg, from {@link
     * #getWriter() getWriter()}{@code .getLength()}).
     *
     * @return the size of this chunk.
     * @since  1.0
     * @see    #setSize(long)
     * @see    #updateSize()
     * @see    Writer#getLength()
     */

    public long getSize( )
    {
        return size;
    }

    //------------------------------------------------------------------

    /**
     * Returns the chunk reader for this chunk.  The reader is used to read the chunk data from a data
     * input, usually a Nested List File document.
     *
     * @return the {@linkplain Reader chunk reader} that will read the data of this chunk.
     * @since  1.0
     * @see    #setReader(Reader)
     * @see    Reader
     */

    public Reader getReader( )
    {
        return reader;
    }

    //------------------------------------------------------------------

    /**
     * Returns the chunk writer for this chunk.  The writer is used to write the chunk data to a data
     * output, usually a Nested List File document.
     *
     * @return the {@linkplain Writer chunk writer} that will write the data of this chunk.
     * @since  1.0
     * @see    #setWriter(Writer)
     * @see    Writer
     */

    public Writer getWriter( )
    {
        return writer;
    }

    //------------------------------------------------------------------

    /**
     * Returns the chunk encoder for this chunk.  The encoder is used by {@link #toXml(org.w3c.dom.Document)
     * toXml(org.w3c.dom.Document)} to encode the chunk data for use as a text node of an XML file.
     *
     * @return the {@linkplain Encoder chunk encoder} that will encode the data of this chunk.
     * @since  1.0
     * @see    #setEncoder(Encoder)
     * @see    Encoder
     */

    public Encoder getEncoder( )
    {
        return encoder;
    }

    //------------------------------------------------------------------

    /**
     * Sets the size of this chunk to the specified value.
     *
     * @param size  the value to which the size of this chunk is to be set.
     * @since 1.0
     * @see   #getSize()
     * @see   #updateSize()
     */

    public void setSize( long size )
    {
        this.size = size;
    }

    //------------------------------------------------------------------

    /**
     * Sets the chunk reader for this chunk to the specified reader.  The reader is used to read the chunk
     * data from a data input, usually a Nested List File document.
     *
     * @param reader  the {@link Reader Reader} that is to be this chunk's reader.
     * @since 1.0
     * @see   #getReader()
     */

    public void setReader( Reader reader )
    {
        this.reader = reader;
    }

    //------------------------------------------------------------------

    /**
     * Sets the chunk writer for this chunk to the specified writer.  The writer is used to write the chunk
     * data to a data output, usually a Nested List File document.
     *
     * @param writer  the {@link Writer Writer} that is to be this chunk's writer.
     * @since 1.0
     * @see   #getWriter()
     */

    public void setWriter( Writer writer )
    {
        this.writer = writer;
    }

    //------------------------------------------------------------------

    /**
     * Sets the chunk encoder for this chunk to the specified encoder.  The encoder is used by {@link
     * #toXml(org.w3c.dom.Document) toXml(org.w3c.dom.Document)} to encode the chunk data for use as a text
     * node of an XML file.
     *
     * @param encoder  the {@link Encoder Encoder} that is to be this chunk's encoder.
     * @since 1.0
     * @see   #getEncoder()
     */

    public void setEncoder( Encoder encoder )
    {
        this.encoder = encoder;
    }

    //------------------------------------------------------------------

    /**
     * Tests whether this chunk is an {@linkplain Attributes attributes chunk} or a subclass of an
     * attributes chunk.
     *
     * @return {@code true} if this chunk is an attributes chunk or a subclass of an attributes chunk;
     *         {@code false} otherwise.
     * @since  1.0
     */

    public boolean isAttributes( )
    {
        return ( this instanceof Attributes );
    }

    //------------------------------------------------------------------

    /**
     * Tests whether this chunk is a {@linkplain ChunkList chunk list} or a subclass of a chunk list.
     *
     * @return {@code true} if this chunk is a chunk list or a subclass of a chunk list; {@code false}
     *         otherwise.
     * @since  1.0
     */

    public boolean isList( )
    {
        return ( this instanceof ChunkList );
    }

    //------------------------------------------------------------------

    /**
     * Updates the chunk's {@code size} field from the chunk writer, if one has been set.  The chunk size is
     * set to the value returned by the chunk writer's {@link Writer#getLength() getLength()} method.
     *
     * @since 1.0
     * @see   #getSize()
     * @see   #setSize(long)
     * @see   Writer#getLength()
     */

    public void updateSize( )
    {
        if ( writer != null )
            size = writer.getLength( );
    }

    //------------------------------------------------------------------

    /**
     * Returns the namespace name that applies to this chunk.  A chunk's namespace name is the first
     * non-{@code null} local namespace name that is encountered in an ancestor when ascending the document
     * tree.  If none of the chunk's ancestors has a local namespace name, {@code null} is returned.
     *
     * @return the namespace name that applies to this chunk, or {@code null} if no namespace name applies.
     * @since  1.0
     * @see    ChunkList#getNamespaceName()
     */

    public String getNamespaceName( )
    {
        String namespaceName = null;
        ChunkList list = parent;
        while ( list != null )
        {
            namespaceName = list.getLocalNamespaceName( );
            if ( namespaceName != null )
                break;
            list = list.parent;
        }
        return namespaceName;
    }

    //------------------------------------------------------------------

    /**
     * Returns the name of this chunk.  A chunk's name is formed by prefixing its namespace name and a
     * vertical line character (U+007C) to its identifier.  If the chunk has no namespace name, its name is
     * just its identifier.
     *
     * @return the name of this chunk.
     * @since  1.0
     * @see    ChunkList#getName()
     */

    public String getName( )
    {
        return id.toName( getNamespaceName( ) );
    }

    //------------------------------------------------------------------

    /**
     * Returns the pathname of this chunk.  A pathname, which might not be unique within a particular
     * document, is formed by concatenating the list-instance identifiers of the ancestors of this chunk
     * with the identifier of this chunk.  The components of the pathname (ie, the identifiers) are
     * separated with '.' (U+002E).  The pathname has a leading '.' if its first component is not the root
     * list.
     *
     * @return the pathname of this chunk.
     * @since  1.0
     */

    public String getPathname( )
    {
        String prefix = (parent == null) ? new String( ) : parent.getPathname( );
        return ( prefix + NlfConstants.PATHNAME_SEPARATOR_CHAR + id.getValue( ) );
    }

    //------------------------------------------------------------------

    /**
     * Returns the size of the chunk header (identifier and size).
     *
     * @return the size of the chunk header (identifier and size).
     * @since  1.0
     */

    public int getHeaderSize( )
    {
        return ( id.getFieldSize( ) + SIZE_SIZE );
    }

    //------------------------------------------------------------------

    /**
     * Returns the size of the chunk as an array of bytes, with the specified byte order.
     *
     * @param  littleEndian  {@code true} if the byte order is little-endian; {@code false} if the byte
     *                       order is big-endian.
     * @return the size of the chunk as an array of bytes.
     * @since  1.0
     */

    public byte[] getSizeBytes( boolean littleEndian )
    {
        byte[] buffer = new byte[SIZE_SIZE];
        Util.longToBytes( size, buffer, 0, buffer.length, littleEndian );
        return buffer;
    }

    //------------------------------------------------------------------

    /**
     * Writes the chunk header (identifier and size) to the specified data output.
     *
     * @param  dataOutput  the data output to which the chunk header is to be written.
     * @throws IOException
     *           if an I/O error occurs.
     * @since  1.0
     */

    public void writeHeader( DataOutput dataOutput )
        throws IOException
    {
        // Write identifier
        id.write( dataOutput );

        // Write size
        dataOutput.write( getSizeBytes( document.isLittleEndian( ) ) );
    }

    //------------------------------------------------------------------

    /**
     * Returns this chunk as an XML element of the specified document.  The chunk data are obtained from
     * the chunk's reader, and encoded as text with the chunk's encoder.  The default encoder is the single
     * instance of {@link Base64Encoder Base64Encoder}.
     *
     * @param  xmlDocument  the XML document that is to be the owner of the element that is created.
     * @return the XML element that is generated from this chunk.
     * @throws DOMException
     *           if an exception occurs in creating the XML element.
     * @throws IOException
     *           if an I/O error occurs in reading the chunk data.
     * @throws NlfException
     *           if
     *           <ul>
     *             <li>an exception occurs in encoding the chunk data, or</li>
     *             <li>there is not enough memory to create an XML element.</li>
     *           </ul>
     * @since  1.0
     */

    public Element toXml( org.w3c.dom.Document xmlDocument )
        throws DOMException, IOException, NlfException
    {
        final   int DEFAULT_INPUT_LENGTH    = 1024;

        Element element = xmlDocument.createElement( id.getValue( ) );
        if ( (size > 0) && (reader != null) && (encoder != null) )
        {
            try
            {
                // Reset reader and encoder
                reader.reset( );
                encoder.reset( );

                // Allocate buffer for block of chunk data
                int inLength = encoder.getInputLength( size );
                if ( inLength == 0 )
                    inLength = DEFAULT_INPUT_LENGTH;
                byte[] buffer = new byte[inLength];

                // Read and encode successive blocks of chunk data, appending each block to the XML element
                // as a text node, and merging each appended text node with previous character data
                long remainingSize = size;
                while ( remainingSize > 0 )
                {
                    int length = (int)Math.min( remainingSize, buffer.length );
                    reader.getDataInput( ).readFully( buffer, 0, length );
                    try
                    {
                        String text = encoder.encode( buffer, 0, length, remainingSize == length );
                        element.appendChild( xmlDocument.createTextNode( text ) );
                    }
                    catch ( IllegalArgumentException e )
                    {
                        throw new NlfException( ExceptionId.ERROR_ENCODING_CHUNK_DATA );
                    }
                    element.normalize( );
                    remainingSize -= length;
                }
            }
            catch ( OutOfMemoryError e )
            {
                throw new NlfException( ExceptionId.NOT_ENOUGH_MEMORY );
            }
        }
        return element;
    }

    //------------------------------------------------------------------

    /**
     * Sets this chunk's parent to the specified list.
     *
     * @param list  the list that is to be this chunk's parent.
     * @since 1.0
     */

    protected void setParent( ChunkList list )
    {
        parent = list;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    protected   Document    document;
    protected   ChunkList   parent;
    protected   Id          id;
    protected   long        size;
    protected   Reader      reader;
    protected   Writer      writer;
    protected   Encoder     encoder;

}

//----------------------------------------------------------------------
