/*====================================================================*\

Id.java

Nested List File identifier class.

------------------------------------------------------------------------

This file is part of the Blank Aspect NLF library (Java package
uk.org.blankaspect.nlf).
Copyright 2007-2013 Andy Morgan-Richards.

The Blank Aspect NLF library is free software: you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package uk.org.blankaspect.nlf;

//----------------------------------------------------------------------


// IMPORTS


import java.io.DataOutput;
import java.io.IOException;

//----------------------------------------------------------------------


// NESTED LIST FILE IDENTIFIER CLASS


/**
 * This class implements an identifier that is used to identify the chunks in a Nested List File.  An
 * identifier is a string that is encoded as a size byte followed by a UTF-8 sequence.  The size of the
 * UTF-8 sequence must be between 1 and 255 bytes.
 * <p>
 * Identifiers that start with the '$' character (U+0024) are reserved.
 * </p>
 * <p>
 * To allow the conversion of a Nested List File to XML, a non-reserved identifer must be a valid unprefixed
 * name (ie, a name that doesn't contain a ':') under XML 1.1.  Note that XML 1.1 names are less restrictive
 * than those of XML 1.0, so that a name that is valid under XML 1.1 might not be valid under XML 1.0.
 * </p>
 *
 * @since 1.0
 */

public class Id
    implements Comparable<Id>
{

////////////////////////////////////////////////////////////////////////
//  Constants
////////////////////////////////////////////////////////////////////////

    // Size of identifier
    public static final     int SIZE_SIZE   = 1;
    public static final     int MIN_SIZE    = 1;
    public static final     int MAX_SIZE    = 255;
    protected static final  int SIZE_MASK   = (1 << (SIZE_SIZE << 3)) - 1;

    // Prefix of reserved identifiers
    public static final     char    RESERVED_PREFIX_CHAR    = '$';
    public static final     String  RESERVED_PREFIX         = Character.toString( RESERVED_PREFIX_CHAR );

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    /**
     * Constructs an identifier from an array of bytes.  The bytes are expected to be in the form in which
     * an identifier appears in a Nested List File: the first byte is the length of the UTF-8 sequence that
     * follows.  The UTF-8 sequence is an encoding of the value of the identifier.
     *
     * @param  bytes  the array of bytes from which the identifier is constructed.  The first byte of the
     *                array is the size of the UTF-8 sequence that follows.
     * @throws IllegalArgumentException
     *           if
     *           <ul>
     *             <li>the first byte of {@code bytes} is zero, or</li>
     *             <li>the bytes starting at {@code bytes[1]} are not a valid UTF-8 sequence, or</li>
     *             <li>the UTF-8 sequence, when decoded, is not a valid identifier.</li>
     *           </ul>
     * @since  1.0
     */

    public Id( byte[] bytes )
        throws IllegalArgumentException
    {
        this( bytes, 0 );
    }

    //------------------------------------------------------------------

    /**
     * Constructs an identifier from an array of bytes, starting at the specified offset.  The bytes are
     * expected to be in the form in which an identifier appears in a Nested List File: the byte at {@code
     * offset} is the length of the UTF-8 sequence that follows.  The UTF-8 sequence is an encoding of the
     * value of the identifier.
     *
     * @param  bytes   the array of bytes from which the identifier is constructed.  The byte of the array
     *                 at {@code offset} is the size of the UTF-8 sequence that follows.
     * @param  offset  the offset to {@code bytes} at which the identifier begins.
     * @throws IllegalArgumentException
     *           if
     *           <ul>
     *             <li>the byte of {@code bytes} at {@code offset} is zero, or</li>
     *             <li>the bytes starting at {@code bytes[offset+1]} are not a valid UTF-8 sequence, or</li>
     *             <li>the UTF-8 sequence, when decoded, is not a valid identifier.</li>
     *           </ul>
     * @since  1.0
     */

    public Id( byte[] bytes,
               int    offset )
        throws IllegalArgumentException
    {
        this( NlfUtilities.utf8ToString( bytes, offset + SIZE_SIZE, getSize( bytes, offset ) ) );
    }

    //------------------------------------------------------------------

    /**
     * Constructs an identifier from a string.
     *
     * @param  str  the string from which the identifier is constructed.
     * @throws IllegalArgumentException
     *           if
     *           <ul>
     *             <li>the string is empty, or</li>
     *             <li>the UTF-8 encoding of the string is longer than 255 bytes, or</li>
     *             <li>the string contains an invalid character.</li>
     *           </ul>
     * @since  1.0
     */

    public Id( String str )
        throws IllegalArgumentException
    {
        // Validate string
        if ( !isValidId( str ) )
            throw new IllegalArgumentException( );

        // Set value
        value = str;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    /**
     * Returns the size of an identifier that is encoded in an array of bytes at a specified offset.
     *
     * @param  bytes   the array of bytes that contains the encoded identifier.
     * @param  offset  the offset to {@code bytes} at which the identifier begins.
     * @return the size of the encoded identifier (ie, the value of the byte of {@code bytes} at {@code
     *         offset}).
     * @since  1.0
     */

    public static int getSize( byte[] bytes,
                               int    offset )
    {
        return ( bytes[offset] & SIZE_MASK );
    }

    //------------------------------------------------------------------

    /**
     * Tests whether the specified string is a valid identifier.  A non-reserved identifier is valid if it
     * is a valid unprefixed name (ie, a name that doesn't contain a ':') under XML 1.1.
     *
     * @param  str  the string that is to be tested.
     * @return {@code true} the specified string is a valid identifier; {@code false} otherwise.
     * @since  1.0
     */

    public static boolean isValidId( String str )
    {
        if ( !NlfUtilities.isUtf8LengthWithinBounds( str, MIN_SIZE, MAX_SIZE ) )
            return false;

        int index = 0;
        while ( index < str.length( ) )
        {
            int codePoint = str.codePointAt( index );
            if ( index == 0 )
            {
                if ( !NlfUtilities.isNameStartChar( codePoint ) && (codePoint != RESERVED_PREFIX_CHAR) )
                    return false;
            }
            else
            {
                if ( !NlfUtilities.isNameChar( codePoint ) )
                    return false;
            }
            index += Character.charCount( codePoint );
        }
        return true;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods : Comparable interface
////////////////////////////////////////////////////////////////////////

    /**
     * Compares this identifier with the specified {@code Id} object.  The comparison is performed by
     * comparing the values of the two objects using {@code java.lang.String.compareTo(java.lang.String)},
     * which compares strings lexicographically by the Unicode value of each character in the strings.
     * <p>
     * The result is
     * </p>
     * <ul>
     *   <li>a negative integer if the value of this identifier lexicographically precedes the value of
     *       the argument;</li>
     *   <li>a positive integer if the value of this identifier lexicographically succeeds the value of
     *       the argument;</li>
     *   <li>zero if the values of the two identifiers are equal.</li>
     * </ul>
     * <p>
     * A result of zero implies that the {@link #equals(Object) equals(Object)} method would return {@code
     * true}.
     * </p>
     * <p>
     * Because of the restrictions on the characters in an identifier, reserved identifiers will always
     * precede non-reserved identifiers.
     * </p>
     *
     * @param  id  the {@code Id} object with which the comparison is to be made.
     * @return <ul>
     *           <li>{@code 0} (zero) if the value of this identifier is equal to the value of {@code
     *               id};</li>
     *           <li>a value less than {@code 0} if the value of this identifier is lexicographically less
     *               than the value of {@code id};</li>
     *           <li>a value greater than {@code 0} if the value of this identifier is lexicographically
     *               greater than the value of {@code id}.</li>
     *         </ul>
     * @since  1.0
     */

    public int compareTo( Id id )
    {
        return value.compareTo( id.value );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods : overriding methods
////////////////////////////////////////////////////////////////////////

    /**
     * Compares this identifier with the specified object.  The result is {@code true} if and only if {@code
     * obj} is an {@code Id} object that has the same value as this identifier.
     *
     * @param  obj  the object with which the comparison is to be made.
     * @return {@code true} if the objects have the same value; {@code false} otherwise.
     * @since  1.0
     */

    @Override
    public boolean equals( Object obj )
    {
        return ( (obj instanceof Id) && value.equals( ((Id)obj).value ) );
    }

    //------------------------------------------------------------------

    /**
     * Returns the hash code for this object.
     *
     * @return the hash code for this object.
     * @since  1.1
     * @see    #equals(Object)
     */

    @Override
    public int hashCode( )
    {
        return value.hashCode( );
    }

    //------------------------------------------------------------------

    /**
     * Returns the identifier as a string.  The string is just the value of the identifier (ie, the value
     * returned by {@link #getValue() getValue()} ).
     *
     * @return the identifier as a string.
     * @since  1.0
     * @see    #getValue()
     */

    @Override
    public String toString( )
    {
        return value;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    /**
     * Returns the value of the identifier.
     *
     * @return the value of the identifier.
     * @since  1.0
     */

    public String getValue( )
    {
        return value;
    }

    //------------------------------------------------------------------

    /**
     * Returns the size of the identifier when it is encoded as a UTF-8 sequence.  The size does not include
     * the initial size byte.
     *
     * @return the size of the UTF-8 encoding of identifier.
     * @since  1.0
     * @see    #getFieldSize()
     */

    public int getSize( )
    {
        return NlfUtilities.getUtf8Length( value );
    }

    //------------------------------------------------------------------

    /**
     * Returns the size of the field that the identifier would occupy in a Nested List File.  The size
     * includes the initial size byte and the UTF-8 encoding of the identifier.
     *
     * @return the size of the UTF-8 encoding of the identifier, including the initial size byte.
     * @since  1.0
     * @see    #getSize()
     */

    public int getFieldSize( )
    {
        return ( SIZE_SIZE + getSize( ) );
    }

    //------------------------------------------------------------------

    /**
     * Tests whether an identifier is reserved (ie, whether it starts with the reserved prefix).
     *
     * @return {@code true} if the identifier is reserved; {@code false} otherwise.
     * @since  1.0
     */

    public boolean isReserved( )
    {
        return ( value.startsWith( RESERVED_PREFIX ) );
    }

    //------------------------------------------------------------------

    /**
     * Converts this identifier to a name, using the specified prefix.  The name is formed from the prefix
     * and the value of the identifier: if the prefix is {@code null}, this method returns the string that
     * is returned by {@link #getValue() getValue()}; otherwise, it returns the concatenation of the prefix,
     * a vertical line character (U+007C) and the string that is returned by {@link #getValue() getValue()}.
     *
     * @param  prefix  the prefix that is to be used to form the name.
     * @return a string formed from the specified prefix and the value of this identifier.
     * @since  1.0
     */

    public String toName( String prefix )
    {
        return ( (prefix == null) ? value : prefix + NlfConstants.NAME_SEPARATOR_CHAR + value );
    }

    //------------------------------------------------------------------

    /**
     * Returns the encoded form of the identifier (a size byte followed by a UTF-8 sequence) as an array of
     * bytes.
     *
     * @return a byte array containing the encoded form of the identifier: a size byte followed by a UTF-8
     *         sequence.
     * @since  1.0
     */

    public byte[] getBytes( )
    {
        byte[] bytes = NlfUtilities.stringToUtf8( value );
        byte[] buffer = new byte[SIZE_SIZE + bytes.length];
        buffer[0] = (byte)bytes.length;
        System.arraycopy( bytes, 0, buffer, SIZE_SIZE, bytes.length );
        return buffer;
    }

    //------------------------------------------------------------------

    /**
     * Writes the encoded form of the identifier (a size byte followed by a UTF-8 sequence) to the
     * specified data output (an instance of {@code java.io.DataOuput}).
     *
     * @param  dataOutput  the data output to which the encoded form of the identifier (a size byte followed
     *                     by a UTF-8 sequence) is to be written.
     * @throws IOException
     *           if an I/O error occurs.
     * @see    #getBytes()
     * @since  1.0
     */

    public void write( DataOutput dataOutput )
        throws IOException
    {
        dataOutput.write( getBytes( ) );
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    private String  value;

}

//----------------------------------------------------------------------
