/*====================================================================*\

NlfException.java

Nested List File exception class.

------------------------------------------------------------------------

This file is part of the Blank Aspect NLF library (Java package
uk.org.blankaspect.nlf).
Copyright 2007-2013 Andy Morgan-Richards.

The Blank Aspect NLF library is free software: you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package uk.org.blankaspect.nlf;

//----------------------------------------------------------------------


// IMPORTS


import java.io.File;

//----------------------------------------------------------------------


// NESTED LIST FILE EXCEPTION CLASS


/**
 * An object of this class signals that an exception has occurred during an operation on a Nested List File.
 * An exception has an identifier that maps to a detail message, and may include the file on which the
 * exception occurred and the offset within the file at which it occurred.  The exception may also act as a
 * wrapper for another {@code java.lang.Throwable}, which is known as the <i>cause</i>.
 * <p>
 * The file, offset and cause are not automatically incorporated into the detail message of an {@code
 * NlfException}, but they can be obtained, along with the exception identifier, by methods of this class or
 * a superclass.
 * </p>
 *
 * @since 1.0
 * @see   NlfUncheckedException
 */

public class NlfException
    extends Exception
{

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    /**
     * Constructs an {@code NlfException} with the specified identifier.  The identifier maps to a message
     * that is used as the exception's detail message in the constructor of {@code java.lang.Exception}.
     *
     * @param id  the exception identifier.
     * @since 1.0
     */

    public NlfException( ExceptionId id )
    {
        super( id.getMessage( ) );
        this.id = id;
        offset = -1;
    }

    //------------------------------------------------------------------

    /**
     * Constructs an {@code NlfException} with the specified identifier and cause.  The identifier maps to
     * a message that is used as the exception's detail message in the constructor of {@code
     * java.lang.Exception}.
     *
     * @param id     the exception identifier.
     * @param cause  the {@code java.lang.Throwable} that caused this exception.
     * @since 1.0
     */

    public NlfException( ExceptionId id,
                         Throwable   cause )
    {
        super( id.getMessage( ), cause );
        this.id = id;
        offset = -1;
    }

    //------------------------------------------------------------------

    /**
     * Constructs an {@code NlfException} with the specified identifier and file.  The identifier maps to
     * a message that is used as the exception's detail message in the constructor of {@code
     * java.lang.Exception}.
     *
     * @param id    the exception identifier.
     * @param file  the file that is associated with the exception, usually the file on which the exception
     *              occurred.
     * @since 1.0
     */

    public NlfException( ExceptionId id,
                         File        file )
    {
        this( id, file, -1 );
    }

    //------------------------------------------------------------------

    /**
     * Constructs an {@code NlfException} with the specified identifier, file and cause.  The identifier
     * maps to a message that is used as the exception's detail message in the constructor of {@code
     * java.lang.Exception}.
     *
     * @param id     the exception identifier.
     * @param file   the file that is associated with the exception, usually the file on which the exception
     *               occurred.
     * @param cause  the {@code java.lang.Throwable} that caused this exception.
     * @since 1.0
     */

    public NlfException( ExceptionId id,
                         File        file,
                         Throwable   cause )
    {
        this( id, file, -1, cause );
    }

    //------------------------------------------------------------------

    /**
     * Constructs an {@code NlfException} with the specified identifier, file and offset.  The identifier
     * maps to a message that is used as the exception's detail message in the constructor of {@code
     * java.lang.Exception}.
     *
     * @param id      the exception identifier.
     * @param file    the file that is associated with the exception, usually the file on which the
     *                exception occurred.
     * @param offset  the offset that is associated with the exception, usually the offset to {@code file}
     *                at which the exception occurred.
     * @since 1.0
     */

    public NlfException( ExceptionId id,
                         File        file,
                         long        offset )
    {
        super( id.getMessage( ) );
        this.id = id;
        this.file = file;
        this.offset = (offset < 0) ? -1 : offset;
    }

    //------------------------------------------------------------------

    /**
     * Constructs an {@code NlfException} with the specified identifier, file, offset and cause.  The
     * identifier maps to a message that is used as the exception's detail message in the constructor of
     * {@code java.lang.Exception}.
     *
     * @param id      the exception identifier.
     * @param file    the file that is associated with the exception, usually the file on which the
     *                exception occurred.
     * @param offset  the offset that is associated with the exception, usually the offset to {@code file}
     *                at which the exception occurred.
     * @param cause   the {@code java.lang.Throwable} that caused this exception.
     * @since 1.0
     */

    public NlfException( ExceptionId id,
                         File        file,
                         long        offset,
                         Throwable   cause )
    {
        super( id.getMessage( ), cause );
        this.id = id;
        this.file = file;
        this.offset = (offset < 0) ? -1 : offset;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance methods
////////////////////////////////////////////////////////////////////////

    /**
     * Returns the exception identifier.
     *
     * @return the exception identifier.
     * @since  1.0
     */

    public ExceptionId getId( )
    {
        return id;
    }

    //------------------------------------------------------------------

    /**
     * Returns the file that is associated with the exception.  The file will usually be the file on which
     * the exception occurred.
     *
     * @return the file that is associated with the exception.  A {@code null} value indicates that no file
     *         is associated with the exception.
     * @since  1.0
     * @see    #getOffset()
     */

    public File getFile( )
    {
        return file;
    }

    //------------------------------------------------------------------

    /**
     * Returns the offset that is associated with the exception, or {@code -1} if the offset is not set to a
     * legal value.  The offset will usually be an offset to the file returned by {@link #getFile()} at
     * which the exception occurred.
     *
     * @return the offset that is associated with the exception.  A value of {@code -1} indicates that the
     *         offset has not been set, or that it was set to an illegal value.
     * @since  1.0
     * @see    #getFile()
     */

    public long getOffset( )
    {
        return offset;
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Instance variables
////////////////////////////////////////////////////////////////////////

    private ExceptionId id;
    private File        file;
    private long        offset;

}

//----------------------------------------------------------------------
