/*====================================================================*\

Util.java

Utility methods class.

------------------------------------------------------------------------

This file is part of the Blank Aspect NLF library (Java package
uk.org.blankaspect.nlf).
Copyright 2007-2013 Andy Morgan-Richards.

The Blank Aspect NLF library is free software: you can redistribute
it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

\*====================================================================*/


// PACKAGE


package uk.org.blankaspect.nlf;

//----------------------------------------------------------------------


// UTILITY METHODS CLASS


/**
 * This class contains a number of utility methods that are used by classes in the Nested List File package.
 *
 * @since 1.0
 */

class Util
{

////////////////////////////////////////////////////////////////////////
//  Constructors
////////////////////////////////////////////////////////////////////////

    private Util( )
    {
    }

    //------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
//  Class methods
////////////////////////////////////////////////////////////////////////

    /**
     * Converts a sequence of bytes with the specified byte order to an {@code int}.
     *
     * @param  bytes         the byte sequence that is to be converted.
     * @param  offset        the offset to {@code bytes} at which the sequence starts.
     * @param  length        the length of the byte sequence.
     * @param  littleEndian  {@code true} if the byte order of the byte sequence is little-endian; {@code
     *                       false} if the byte order is big-endian.
     * @return the {@code int} that results from the conversion of the byte sequence.
     * @since  1.0
     * @see    #intToBytes(int, byte[], int, int, boolean)
     * @see    #bytesToLong(byte[], int, int, boolean)
     */

    protected static int bytesToInt( byte[]  bytes,
                                     int     offset,
                                     int     length,
                                     boolean littleEndian )
    {
        return ( littleEndian ? bytesToIntLE( bytes, offset, length )
                              : bytesToIntBE( bytes, offset, length ) );
    }

    //------------------------------------------------------------------

    /**
     * Converts an {@code int} to a sequence of bytes of the specified length and with the specified byte
     * order.
     *
     * @param value         the value that is to be converted.
     * @param buffer        the byte array that is to contain the byte sequence.
     * @param offset        the offset to {@code buffer} at which the sequence is to start.
     * @param length        the required length of the byte sequence.
     * @param littleEndian  {@code true} if the byte order of the byte sequence is to be little-endian;
     *                      {@code false} if the byte order is to be big-endian.
     * @since 1.0
     * @see   #bytesToInt(byte[], int, int, boolean)
     * @see   #longToBytes(long, byte[], int, int, boolean)
     */

    protected static void intToBytes( int     value,
                                      byte[]  buffer,
                                      int     offset,
                                      int     length,
                                      boolean littleEndian )
    {
        if ( littleEndian )
            intToBytesLE( value, buffer, offset, length );
        else
            intToBytesBE( value, buffer, offset, length );
    }

    //------------------------------------------------------------------

    /**
     * Converts a sequence of bytes with the specified byte order to a {@code long}.
     *
     * @param  bytes         the byte sequence that is to be converted.
     * @param  offset        the offset to {@code bytes} at which the sequence starts.
     * @param  length        the length of the byte sequence.
     * @param  littleEndian  {@code true} if the byte order of the byte sequence is little-endian; {@code
     *                       false} if the byte order is big-endian.
     * @return the {@code long} that results from the conversion of the byte sequence.
     * @since  1.0
     * @see    #longToBytes(long, byte[], int, int, boolean)
     * @see    #bytesToInt(byte[], int, int, boolean)
     */

    protected static long bytesToLong( byte[]  bytes,
                                       int     offset,
                                       int     length,
                                       boolean littleEndian )
    {
        return ( littleEndian ? bytesToLongLE( bytes, offset, length )
                              : bytesToLongBE( bytes, offset, length ) );
    }

    //------------------------------------------------------------------

    /**
     * Converts a {@code long} to a sequence of bytes of the specified length and with the specified byte
     * order.
     *
     * @param value         the value that is to be converted.
     * @param buffer        the byte array that is to contain the byte sequence.
     * @param offset        the offset to {@code buffer} at which the sequence is to start.
     * @param length        the required length of the byte sequence.
     * @param littleEndian  {@code true} if the byte order of the byte sequence is to be little-endian;
     *                      {@code false} if the byte order is to be big-endian.
     * @since 1.0
     * @see   #bytesToLong(byte[], int, int, boolean)
     * @see   #intToBytes(int, byte[], int, int, boolean)
     */

    protected static void longToBytes( long    value,
                                       byte[]  buffer,
                                       int     offset,
                                       int     length,
                                       boolean littleEndian )
    {
        if ( littleEndian )
            longToBytesLE( value, buffer, offset, length );
        else
            longToBytesBE( value, buffer, offset, length );
    }

    //------------------------------------------------------------------

    /**
     * Converts a big-endian sequence of bytes to an {@code int}.
     *
     * @param  bytes   the byte sequence that is to be converted.
     * @param  offset  the offset to {@code bytes} at which the sequence starts.
     * @param  length  the length of the byte sequence.
     * @return the {@code int} that results from the conversion of the byte sequence.
     * @since  1.0
     * @see    #bytesToIntLE(byte[], int, int)
     * @see    #intToBytesBE(int, byte[], int, int)
     */

    private static int bytesToIntBE( byte[] bytes,
                                     int    offset,
                                     int    length )
    {
        int endOffset = offset + length;
        int value = bytes[offset];
        while ( ++offset < endOffset )
        {
            value <<= 8;
            value |= bytes[offset] & 0xFF;
        }
        return value;
    }

    //------------------------------------------------------------------

    /**
     * Converts a little-endian sequence of bytes to an {@code int}.
     *
     * @param  bytes   the byte sequence that is to be converted.
     * @param  offset  the offset to {@code bytes} at which the sequence starts.
     * @param  length  the length of the byte sequence.
     * @return the {@code int} that results from the conversion of the byte sequence.
     * @since  1.0
     * @see    #bytesToIntBE(byte[], int, int)
     * @see    #intToBytesLE(int, byte[], int, int)
     */

    private static int bytesToIntLE( byte[] bytes,
                                     int    offset,
                                     int    length )
    {
        int i = offset + length;
        int value = bytes[--i];
        while ( --i >= offset )
        {
            value <<= 8;
            value |= bytes[i] & 0xFF;
        }
        return value;
    }

    //------------------------------------------------------------------

    /**
     * Converts an {@code int} to a big-endian sequence of bytes of the specified length.
     *
     * @param value   the value that is to be converted.
     * @param buffer  the byte array that is to contain the byte sequence.
     * @param offset  the offset to {@code buffer} at which the sequence is to start.
     * @param length  the required length of the byte sequence.
     * @since 1.0
     * @see   #intToBytesLE(int, byte[], int, int)
     * @see   #bytesToIntBE(byte[], int, int, boolean)
     */

    private static void intToBytesBE( int    value,
                                      byte[] buffer,
                                      int    offset,
                                      int    length )
    {
        for ( int i = offset + length - 1; i >= offset; --i )
        {
            buffer[i] = (byte)value;
            value >>>= 8;
        }
    }

    //------------------------------------------------------------------

    /**
     * Converts an {@code int} to a little-endian sequence of bytes of the specified length.
     *
     * @param value   the value that is to be converted.
     * @param buffer  the byte array that is to contain the byte sequence.
     * @param offset  the offset to {@code buffer} at which the sequence is to start.
     * @param length  the required length of the byte sequence.
     * @since 1.0
     * @see   #intToBytesBE(int, byte[], int, int)
     * @see   #bytesToIntLE(byte[], int, int, boolean)
     */

    private static void intToBytesLE( int    value,
                                      byte[] buffer,
                                      int    offset,
                                      int    length )
    {
        int endOffset = offset + length;
        for ( int i = offset; i < endOffset; ++i )
        {
            buffer[i] = (byte)value;
            value >>>= 8;
        }
    }

    //------------------------------------------------------------------

    /**
     * Converts a big-endian sequence of bytes to a {@code long}.
     *
     * @param  bytes   the byte sequence that is to be converted.
     * @param  offset  the offset to {@code bytes} at which the sequence starts.
     * @param  length  the length of the byte sequence.
     * @return the {@code long} that results from the conversion of the byte sequence.
     * @since  1.0
     * @see    #bytesToLongLE(byte[], int, int)
     * @see    #longToBytesBE(long, byte[], int, int)
     */

    private static long bytesToLongBE( byte[] bytes,
                                       int    offset,
                                       int    length )
    {
        int endOffset = offset + length;
        long value = bytes[offset];
        while ( ++offset < endOffset )
        {
            value <<= 8;
            value |= bytes[offset] & 0xFF;
        }
        return value;
    }

    //------------------------------------------------------------------

    /**
     * Converts a little-endian sequence of bytes to a {@code long}.
     *
     * @param  bytes   the byte sequence that is to be converted.
     * @param  offset  the offset to {@code bytes} at which the sequence starts.
     * @param  length  the length of the byte sequence.
     * @return the {@code long} that results from the conversion of the byte sequence.
     * @since  1.0
     * @see    #bytesToLongBE(byte[], int, int)
     * @see    #longToBytesLE(long, byte[], int, int)
     */

    private static long bytesToLongLE( byte[] bytes,
                                       int    offset,
                                       int    length )
    {
        int i = offset + length;
        long value = bytes[--i];
        while ( --i >= offset )
        {
            value <<= 8;
            value |= bytes[i] & 0xFF;
        }
        return value;
    }

    //------------------------------------------------------------------

    /**
     * Converts a {@code long} to a big-endian sequence of bytes of the specified length.
     *
     * @param value   the value that is to be converted.
     * @param buffer  the byte array that is to contain the byte sequence.
     * @param offset  the offset to {@code buffer} at which the sequence is to start.
     * @param length  the required length of the byte sequence.
     * @since 1.0
     * @see   #longToBytesLE(long, byte[], int, int)
     * @see   #bytesToLongBE(byte[], int, int, boolean)
     */

    private static void longToBytesBE( long   value,
                                       byte[] buffer,
                                       int    offset,
                                       int    length )
    {
        for ( int i = offset + length - 1; i >= offset; --i )
        {
            buffer[i] = (byte)value;
            value >>>= 8;
        }
    }

    //------------------------------------------------------------------

    /**
     * Converts a {@code long} to a little-endian sequence of bytes of the specified length.
     *
     * @param value   the value that is to be converted.
     * @param buffer  the byte array that is to contain the byte sequence.
     * @param offset  the offset to {@code buffer} at which the sequence is to start.
     * @param length  the required length of the byte sequence.
     * @since 1.0
     * @see   #longToBytesBE(long, byte[], int, int)
     * @see   #bytesToLongLE(byte[], int, int, boolean)
     */

    private static void longToBytesLE( long   value,
                                       byte[] buffer,
                                       int    offset,
                                       int    length )
    {
        int endOffset = offset + length;
        for ( int i = offset; i < endOffset; ++i )
        {
            buffer[i] = (byte)value;
            value >>>= 8;
        }
    }

    //------------------------------------------------------------------

}

//----------------------------------------------------------------------
